from typing import Dict, TYPE_CHECKING, Any, Sequence, Union

if TYPE_CHECKING:
    from mysql_mimic.results import AllowedResult, Column
    from mysql_mimic.connection import Connection


class Session:
    """
    Abstract client session.

    This should be implemented by applications.
    """

    async def query(
        self, sql: str, attrs: Dict[str, str]
    ) -> "AllowedResult":  # pylint: disable=unused-argument
        """
        Process a SQL query.

        Args:
            sql: SQL statement from client
            attrs: Arbitrary query attributes set by client
        Returns:
            One of:
            - tuple(rows, column_names), where "rows" is a sequence of sequences
              and "column_names" is a sequence of strings with the same length
              as every row in "rows"
            - tuple(rows, result_columns), where "rows" is the same
              as above, and "result_columns" is a sequence of mysql_mimic.ResultColumn
              instances.
            - mysql_mimic.ResultSet instance
        """
        return [], []

    async def close(self) -> None:
        """
        Close the session.

        This is called when the client closes the connection
        """

    async def init(self, connection: "Connection") -> None:
        """
        Called when connection phase is complete.

        This is also called after a COM_CHANGE_USER command completes.

        Args:
            connection: connection of the session
        """

    async def set(self, **kwargs: Dict[str, Any]) -> None:
        """
        Set session variables.

        Args:
            **kwargs: mapping of variable names to values
        """

    async def show_columns(
        self, database: str, table: str
    ) -> Sequence[Union["Column", dict]]:  # pylint: disable=unused-argument
        """
        Show column metadata.

        Args:
            database: database name
            table: table name
        Returns:
            columns
        """
        return []

    async def show_tables(
        self, database: str
    ) -> Sequence[str]:  # pylint: disable=unused-argument
        """
        Show table metadata.

        Args:
            database: database name
        Returns:
            table names
        """
        return []

    async def show_databases(self) -> Sequence[str]:
        """
        Show database metadata.

        Returns:
            database names
        """
        return []

    async def rollback(self) -> None:
        """
        Roll back the current transaction, canceling its changes.
        """



import logging
from ssl import SSLContext
from typing import Optional, Dict, Any, Iterator

from mysql_mimic.auth import (
    AuthInfo,
    Forbidden,
    Success,
    AuthPlugin,
    AuthState,
    IdentityProvider,
)
from mysql_mimic.charset import CharacterSet
from mysql_mimic.constants import DEFAULT_SERVER_CAPABILITIES
from mysql_mimic.errors import ErrorCode, MysqlError
from mysql_mimic.packets import SSLRequest
from mysql_mimic.prepared import PreparedStatement, REGEX_PARAM
from mysql_mimic.results import ensure_result_set, ResultSet
from mysql_mimic import types, packets
from mysql_mimic.session import Session
from mysql_mimic.stream import MysqlStream, ConnectionClosed
from mysql_mimic.types import Capabilities
from mysql_mimic.utils import seq
from mysql_mimic.admin import Admin
from mysql_mimic.variables import SystemVariables

logger = logging.getLogger(__name__)


class Connection:
    _MAX_PREPARED_STMT_ID = 2**32

    def __init__(
        self,
        stream: MysqlStream,
        connection_id: int,
        session: Session,
        identity_provider: IdentityProvider,
        server_capabilities: Capabilities = DEFAULT_SERVER_CAPABILITIES,
        ssl: Optional[SSLContext] = None,
    ):
        self.stream = stream
        self.session = session
        self.connection_id = connection_id
        self.identity_provider = identity_provider
        self.ssl = ssl

        # Authentication plugins can reuse the initial handshake data.
        # This let's clients reuse the nonce when performing COM_CHANGE_USER, skipping a round trip.
        self.handshake_auth_data: Optional[bytes] = None
        self.handshake_auth_plugin: Optional[str] = None

        self.server_capabilities = server_capabilities
        if ssl:
            self.server_capabilities |= Capabilities.CLIENT_SSL

        self.capabilities = Capabilities(0)
        self.status_flags = types.ServerStatus(0)

        self.max_packet_size = 0
        self.client_plugin_name: Optional[str] = None
        self.client_connect_attrs: Dict[str, str] = {}
        self.zstd_compression_level = 0

        self.prepared_stmt_seq = seq(self._MAX_PREPARED_STMT_ID)
        self.prepared_stmts: Dict[int, PreparedStatement] = {}

        self.vars = SystemVariables()
        self.admin = Admin(
            connection_id=connection_id, session=session, variables=self.vars
        )

    @property
    def server_charset(self) -> CharacterSet:
        return self.vars.server_charset

    @property
    def client_charset(self) -> CharacterSet:
        return self.vars.client_charset

    @property
    def database(self) -> Optional[str]:
        return self.admin.database

    @property
    def username(self) -> Optional[str]:
        return self.admin.username

    @username.setter
    def username(self, username: str) -> None:
        self.admin.username = username

    async def start(self) -> None:
        logger.info("Started new connection: %s", self.connection_id)
        try:
            await self.connection_phase()
            await self.session.init(self)
        except Exception as e:
            await self.stream.write(self.error(msg=e, code=ErrorCode.HANDSHAKE_ERROR))
            raise

        try:
            await self.command_phase()
        finally:
            await self.session.close()

    async def connection_phase(self) -> None:
        default_auth_plugin = self.identity_provider.get_default_plugin()
        auth_data, auth_state = await default_auth_plugin.start()
        assert isinstance(auth_data, bytes)
        self.handshake_auth_data = auth_data
        self.handshake_auth_plugin = default_auth_plugin.name

        handshake_v10 = packets.make_handshake_v10(
            capabilities=self.server_capabilities,
            server_charset=self.server_charset,
            server_version=self.vars.mysql_version,
            connection_id=self.connection_id,
            auth_data=auth_data,
            status_flags=self.status_flags,
            auth_plugin_name=default_auth_plugin.name,
        )
        await self.stream.write(handshake_v10)
        response = packets.parse_handshake_response(
            capabilities=self.server_capabilities,
            data=await self.stream.read(),
        )
        if isinstance(response, SSLRequest):
            assert self.ssl is not None
            await self.stream.start_tls(self.ssl)
            response = packets.parse_handshake_response_41(
                capabilities=self.server_capabilities,
                data=await self.stream.read(),
            )

        self.capabilities = response.capabilities
        self.max_packet_size = response.max_packet_size
        self.vars.client_charset = response.client_charset
        self.admin.database = response.database
        self.client_plugin_name = response.client_plugin
        self.client_connect_attrs = response.connect_attrs
        self.zstd_compression_level = response.zstd_compression_level
        self.vars.external_user = response.username

        await self.authenticate(
            auth_state=auth_state,
            server_plugin=default_auth_plugin,
            username=response.username,
            client_plugin_name=response.client_plugin,
            auth_response=response.auth_response,
            connect_attrs=response.connect_attrs,
        )
        self.stream.reset_seq()

    async def handle_change_user(self, data: bytes) -> None:
        com_change_user = packets.parse_com_change_user(
            capabilities=self.capabilities,
            client_charset=self.client_charset,
            data=data,
        )

        self.admin.database = com_change_user.database
        self.vars.external_user = com_change_user.username
        if com_change_user.client_charset:
            self.vars.client_charset = com_change_user.client_charset
        if com_change_user.connect_attrs:
            self.client_connect_attrs = com_change_user.connect_attrs

        await self.authenticate(
            username=com_change_user.username,
            auth_response=com_change_user.auth_response,
            client_plugin_name=com_change_user.client_plugin,
            connect_attrs=com_change_user.connect_attrs,
        )

        await self.session.init(self)

    async def authenticate(
        self,
        username: str,
        auth_response: bytes,
        client_plugin_name: Optional[str],
        connect_attrs: Dict[str, str],
        auth_state: Optional[AuthState] = None,
        server_plugin: Optional[AuthPlugin] = None,
    ) -> None:
        user = await self.identity_provider.get_user(username)

        if not user:
            await self.stream.write(
                self.error(
                    msg=f"User {username} does not exist",
                    code=ErrorCode.USER_DOES_NOT_EXIST,
                )
            )
            return

        user_plugin = (
            self.identity_provider.get_plugin(user.auth_plugin or "")
            or self.identity_provider.get_default_plugin()
        )
        assert self.handshake_auth_plugin is not None
        auth_info = AuthInfo(
            username=username,
            data=auth_response,
            connect_attrs=connect_attrs,
            user=user,
            client_plugin_name=client_plugin_name,
            handshake_auth_data=self.handshake_auth_data,
            handshake_plugin_name=self.handshake_auth_plugin,
        )

        if (
            server_plugin
            and (
                server_plugin.client_plugin_name is None
                or server_plugin.client_plugin_name == client_plugin_name
            )
            and server_plugin.name == user_plugin.name
        ):
            # Optimistic match during handshake
            assert auth_state is not None
            decision = await auth_state.asend(auth_info)
        elif (
            user_plugin.client_plugin_name is None
            or user_plugin.client_plugin_name == client_plugin_name
        ):
            # Continue with provided client plugin
            decision, auth_state = await user_plugin.start(auth_info)
        else:
            # Mismatch - switch authentication method
            decision, auth_state = await user_plugin.start()
            if user_plugin.client_plugin_name and isinstance(decision, bytes):
                await self.stream.write(
                    packets.make_auth_switch_request(
                        server_charset=self.server_charset,
                        plugin_name=user_plugin.client_plugin_name,
                        plugin_provided_data=decision,
                    )
                )
                auth_response = await self.stream.read()
                auth_info = auth_info.copy(auth_response)
                decision = await auth_state.asend(auth_info)

        while not isinstance(decision, (Success, Forbidden)):
            await self.stream.write(packets.make_auth_more_data(decision))
            auth_response = await self.stream.read()
            auth_info = auth_info.copy(auth_response)
            decision = await auth_state.asend(auth_info)

        if isinstance(decision, Success):
            self.admin.username = decision.authenticated_as
            await self.stream.write(self.ok())
        else:
            await self.stream.write(
                self.error(
                    msg=decision.msg or f"Access denied for user {auth_info.user.name}",
                    code=ErrorCode.ACCESS_DENIED_ERROR,
                )
            )

    async def command_phase(self) -> None:
        """https://dev.mysql.com/doc/internals/en/command-phase.html"""
        while True:
            try:
                data = await self.stream.read()
            except ConnectionClosed:
                logger.info("Connection closed")
                return
            try:
                command = data[0]
                rest = data[1:]

                if command == types.Commands.COM_QUERY:
                    await self.handle_query(rest)
                elif command == types.Commands.COM_STMT_PREPARE:
                    await self.handle_stmt_prepare(rest)
                elif command == types.Commands.COM_STMT_SEND_LONG_DATA:
                    await self.handle_stmt_send_long_data(rest)
                elif command == types.Commands.COM_STMT_EXECUTE:
                    await self.handle_stmt_execute(rest)
                elif command == types.Commands.COM_STMT_FETCH:
                    await self.handle_stmt_fetch(rest)
                elif command == types.Commands.COM_STMT_RESET:
                    await self.handle_stmt_reset(rest)
                elif command == types.Commands.COM_STMT_CLOSE:
                    await self.handle_stmt_close(rest)
                elif command == types.Commands.COM_PING:
                    await self.handle_ping(rest)
                elif command == types.Commands.COM_CHANGE_USER:
                    await self.handle_change_user(rest)
                elif command == types.Commands.COM_RESET_CONNECTION:
                    await self.handle_reset_connection(rest)
                elif command == types.Commands.COM_DEBUG:
                    await self.handle_debug(rest)
                elif command == types.Commands.COM_QUIT:
                    return
                else:
                    raise MysqlError(
                        f"Unsupported Command: {hex(command)}",
                        ErrorCode.UNKNOWN_COM_ERROR,
                    )

            except MysqlError as e:
                logger.exception(e)
                await self.stream.write(self.error(msg=e.msg, code=e.code))
            except Exception as e:  # pylint: disable=broad-except
                logger.exception(e)
                await self.stream.write(self.error(msg=e))
            finally:
                self.stream.reset_seq()

    async def handle_ping(self, data: bytes) -> None:  # pylint: disable=unused-argument
        """
        https://dev.mysql.com/doc/internals/en/com-ping.html

        COM_PING check if the server is alive.
        """
        await self.stream.write(self.ok())

    async def handle_reset_connection(
        self, data: bytes
    ) -> None:  # pylint: disable=unused-argument
        """
        https://dev.mysql.com/doc/internals/en/com-reset-connection.html

        COM_RESET_CONNECTION Resets the session state.

        For now, we're just treating this like a no-op.
        """
        await self.stream.write(self.ok())

    async def handle_debug(
        self, data: bytes
    ) -> None:  # pylint: disable=unused-argument
        """
        https://dev.mysql.com/doc/internals/en/com-debug.html

        COM_DEBUG triggers a dump on internal debug info to stdout of the mysql-server.

        For now, we're just treating this like a no-op.
        """
        await self.stream.write(self.ok())

    async def handle_query(self, data: bytes) -> None:
        """
        https://dev.mysql.com/doc/internals/en/com-query.html

        COM_QUERY is used to send the server a text-based query that is executed immediately.
        """

        com_query = packets.parse_com_query(
            capabilities=self.capabilities,
            client_charset=self.client_charset,
            data=data,
        )

        result_set = await self.query(com_query.sql, com_query.query_attrs)

        if not result_set:
            await self.stream.write(self.ok())
            return

        for packet in self.text_resultset(result_set):
            await self.stream.write(packet)

    async def handle_stmt_prepare(self, data: bytes) -> None:
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-prepare.html

        COM_STMT_PREPARE creates a prepared statement from the passed query string.
        """
        sql = self.client_charset.decode(data)

        stmt_id = next(self.prepared_stmt_seq)
        num_params = len(REGEX_PARAM.findall(sql))

        stmt = PreparedStatement(
            stmt_id=stmt_id,
            sql=sql,
            num_params=num_params,
        )
        self.prepared_stmts[stmt_id] = stmt

        for packet in self.com_stmt_prepare_response(stmt):
            await self.stream.write(packet)

    async def handle_stmt_send_long_data(self, data: bytes) -> None:
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-send-long-data.html

        COM_STMT_SEND_LONG_DATA sends the data for a column.
        """
        com_stmt_send_long_data = packets.parse_com_stmt_send_long_data(data)
        stmt = self.get_stmt(com_stmt_send_long_data.stmt_id)
        if stmt.param_buffers is None:
            stmt.param_buffers = {}
        buffer = stmt.param_buffers.setdefault(
            com_stmt_send_long_data.param_id, bytearray()
        )
        buffer.extend(com_stmt_send_long_data.data)

    async def handle_stmt_execute(self, data: bytes) -> None:
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-execute.html

        COM_STMT_EXECUTE asks the server to execute a prepared statement as identified by stmt-id.
        """
        com_stmt_execute = packets.parse_com_stmt_execute(
            capabilities=self.capabilities,
            client_charset=self.client_charset,
            data=data,
            get_stmt=self.get_stmt,
        )

        com_stmt_execute.stmt.param_buffers = None

        result_set = await self.query(
            com_stmt_execute.sql, com_stmt_execute.query_attrs
        )

        if not result_set:
            await self.stream.write(self.ok())
            return

        await self.stream.write(types.uint_len(len(result_set.columns)))

        for column in result_set.columns:
            await self.stream.write(
                packets.make_column_definition_41(
                    server_charset=self.server_charset,
                    name=column.name,
                    column_type=column.type,
                    character_set=column.character_set,
                )
            )

        rows = (
            packets.make_binary_resultrow(r, result_set.columns)
            for r in result_set.rows
        )

        if com_stmt_execute.use_cursor:
            com_stmt_execute.stmt.cursor = rows
            await self.stream.write(
                self.ok_or_eof(flags=types.ServerStatus.SERVER_STATUS_CURSOR_EXISTS)
            )
        else:
            if not self.deprecate_eof():
                await self.stream.write(self.eof())
            for row in rows:
                await self.stream.write(row)
            await self.stream.write(self.ok_or_eof())

    async def handle_stmt_fetch(self, data: bytes) -> None:
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-fetch.html

        COM_STMT_FETCH fetches rows from an existing resultset after a COM_STMT_EXECUTE.
        """
        com_stmt_fetch = packets.parse_handle_stmt_fetch(data)

        stmt = self.get_stmt(com_stmt_fetch.stmt_id)
        assert stmt.cursor is not None
        count = 0
        for _, packet in zip(range(com_stmt_fetch.num_rows), stmt.cursor):
            await self.stream.write(packet)
            count += 1

        done = count < com_stmt_fetch.num_rows

        await self.stream.write(
            self.ok_or_eof(
                flags=types.ServerStatus.SERVER_STATUS_LAST_ROW_SENT
                if done
                else types.ServerStatus.SERVER_STATUS_CURSOR_EXISTS
            )
        )

    async def handle_stmt_reset(self, data: bytes) -> None:
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-reset.html

        COM_STMT_RESET resets the data of a prepared statement which was accumulated with COM_STMT_SEND_LONG_DATA
        commands and closes the cursor if it was opened with COM_STMT_EXECUTE.
        """
        com_stmt_reset = packets.parse_com_stmt_reset(data)
        stmt = self.get_stmt(com_stmt_reset.stmt_id)
        stmt.param_buffers = None
        stmt.cursor = None
        await self.stream.write(self.ok())

    async def handle_stmt_close(self, data: bytes) -> None:
        """
        https://dev.mysql.com/doc/internals/en/com-stmt-close.html

        COM_STMT_CLOSE deallocates a prepared statement.
        """
        com_stmt_close = packets.parse_com_stmt_close(data)
        self.prepared_stmts.pop(com_stmt_close.stmt_id, None)

    def get_stmt(self, stmt_id: int) -> PreparedStatement:
        if stmt_id in self.prepared_stmts:
            return self.prepared_stmts[stmt_id]
        raise MysqlError(f"Unknown statement: {stmt_id}", ErrorCode.UNKNOWN_PROCEDURE)

    async def query(self, sql: str, query_attrs: Dict[str, str]) -> ResultSet:
        result_set = await self.admin.parse(sql)

        if result_set is None:
            sql = self.admin.replace_variables(sql)
            result_set = ensure_result_set(await self.session.query(sql, query_attrs))

        return result_set

    def ok(self, **kwargs: Any) -> bytes:
        return packets.make_ok(
            capabilities=self.capabilities,
            status_flags=self.status_flags,
            **kwargs,
        )

    def eof(self, **kwargs: Any) -> bytes:
        return packets.make_eof(
            capabilities=self.capabilities,
            status_flags=self.status_flags,
            **kwargs,
        )

    def ok_or_eof(
        self,
        affected_rows: int = 0,
        last_insert_id: int = 0,
        warnings: int = 0,
        flags: int = 0,
    ) -> bytes:
        if self.deprecate_eof():
            return self.ok(
                eof=True,
                affected_rows=affected_rows,
                last_insert_id=last_insert_id,
                warnings=warnings,
                flags=flags,
            )
        return self.eof(warnings=warnings, flags=flags)

    def error(self, **kwargs: Any) -> bytes:
        return packets.make_error(
            capabilities=self.capabilities, server_charset=self.server_charset, **kwargs
        )

    def deprecate_eof(self) -> bool:
        return Capabilities.CLIENT_DEPRECATE_EOF in self.capabilities

    def text_resultset(self, result_set: ResultSet) -> Iterator[bytes]:
        yield packets.make_column_count(
            capabilities=self.capabilities, column_count=len(result_set.columns)
        )

        for column in result_set.columns:
            yield packets.make_column_definition_41(
                server_charset=self.server_charset,
                name=column.name,
                column_type=column.type,
                character_set=column.character_set,
            )

        if not self.deprecate_eof():
            yield self.eof()

        affected_rows = 0

        for row in result_set.rows:
            affected_rows += 1
            yield packets.make_text_resultset_row(row, result_set.columns)

        yield self.ok_or_eof(affected_rows=affected_rows)

    def com_stmt_prepare_response(
        self, statement: PreparedStatement
    ) -> Iterator[bytes]:
        yield packets.make_com_stmt_prepare_ok(statement)
        if statement.num_params:
            for _ in range(statement.num_params):
                yield packets.make_column_definition_41(
                    server_charset=self.server_charset, name="?"
                )
            yield self.eof()



import asyncio
import functools
import sqlite3
from ssl import SSLContext
from typing import (
    Optional,
    List,
    Dict,
    Any,
    Callable,
    TypeVar,
    Awaitable,
    Sequence,
    AsyncGenerator,
    Type,
    Tuple,
)

import aiomysql
import mysql.connector
import sqlalchemy.engine
from mysql.connector.connection import (
    MySQLCursorPrepared,
    MySQLCursorDict,
    MySQLConnection,
)
from mysql.connector.cursor import MySQLCursor
from sqlalchemy.ext.asyncio import create_async_engine
import pytest
import pytest_asyncio

from mysql_mimic import MysqlServer, Session
from mysql_mimic.auth import (
    User,
    AuthPlugin,
    IdentityProvider,
)
from mysql_mimic.connection import Connection
from mysql_mimic.results import AllowedResult


class PreparedDictCursor(MySQLCursorPrepared):
    def fetchall(self) -> Optional[List[Dict[str, Any]]]:
        rows = super().fetchall()

        if rows is not None:
            return [dict(zip(self.column_names, row)) for row in rows]

        return None


class MockSession(Session):
    def __init__(self) -> None:
        super().__init__()
        self.return_value: Any = None
        self.echo = False
        self.sqlite = sqlite3.connect(":memory:")
        self.use_sqlite = False
        self.connection: Optional[Connection] = None
        self.last_query_attrs: Optional[Dict[str, str]] = None
        self.users: Optional[Dict[str, User]] = None
        self.columns: Dict[Tuple[str, str], List[dict]] = {}

    async def init(self, connection: Connection) -> None:
        self.connection = connection

    async def query(self, sql: str, attrs: Dict[str, str]) -> AllowedResult:
        self.last_query_attrs = attrs
        if self.use_sqlite:
            cursor = self.sqlite.execute(sql)
            return cursor.fetchall(), [d[0] for d in cursor.description]
        if self.echo:
            return [(sql,)], ["sql"]
        return self.return_value

    async def show_columns(self, database: str, table: str) -> Sequence[dict]:
        return self.columns.get((database, table), [])

    async def show_tables(self, database: str) -> Sequence[str]:
        return [table for db, table in self.columns if db == database]

    async def show_databases(self) -> Sequence[str]:
        return [db for db, _ in self.columns]


class MockIdentityProvider(IdentityProvider):
    def __init__(self, auth_plugins: List[AuthPlugin], users: Dict[str, User]):
        self.auth_plugins = auth_plugins
        self.users = users

    def get_plugins(self) -> Sequence[AuthPlugin]:
        return self.auth_plugins

    async def get_user(self, username: str) -> Optional[User]:
        return self.users.get(username)


T = TypeVar("T")


async def to_thread(func: Callable[..., T], *args: Any, **kwargs: Any) -> T:
    loop = asyncio.get_running_loop()
    func_call = functools.partial(func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)


@pytest.fixture
def session() -> MockSession:
    return MockSession()


@pytest.fixture
def auth_plugins() -> Optional[List[AuthPlugin]]:
    return None


@pytest.fixture
def users() -> Dict[str, User]:
    return {}


@pytest.fixture
def identity_provider(
    auth_plugins: Optional[List[AuthPlugin]], users: Dict[str, User]
) -> Optional[MockIdentityProvider]:
    if auth_plugins:
        return MockIdentityProvider(auth_plugins, users)
    return None


@pytest.fixture
def ssl() -> Optional[SSLContext]:
    return None


@pytest_asyncio.fixture
async def server(
    session: MockSession,
    identity_provider: Optional[MockIdentityProvider],
    ssl: Optional[SSLContext],
) -> AsyncGenerator[MysqlServer, None]:
    srv = MysqlServer(
        session_factory=lambda: session,
        identity_provider=identity_provider,
        ssl=ssl,
    )
    try:
        await srv.start_server(port=3307)
    except OSError as e:
        if e.errno == 48:
            # Port already in use.
            # This should only happen if there is a race condition between concurrent test runs.
            # Getting a new free port can be slow, so we optimistically try to use the free_port fixture first.
            await srv.start_server(port=0)
        else:
            raise
    asyncio.create_task(srv.serve_forever())
    try:
        yield srv
    finally:
        srv.close()
        await srv.wait_closed()


@pytest.fixture
def port(server: MysqlServer) -> int:
    return server.sockets()[0].getsockname()[1]


ConnectFixture = Callable[..., Awaitable[MySQLConnection]]


@pytest.fixture
def connect(port: int) -> ConnectFixture:
    async def conn(**kwargs: Any) -> MySQLConnection:
        return await to_thread(
            mysql.connector.connect, use_pure=True, port=port, **kwargs
        )

    return conn


@pytest_asyncio.fixture
async def mysql_connector_conn(connect: ConnectFixture) -> MySQLConnection:
    conn = await connect()
    try:
        yield conn
    finally:
        conn.close()


@pytest_asyncio.fixture
async def aiomysql_conn(port: int) -> aiomysql.Connection:
    async with aiomysql.connect(port=port) as conn:
        yield conn


@pytest_asyncio.fixture
async def sqlalchemy_engine(port: int) -> sqlalchemy.engine.Engine:
    engine = create_async_engine(url=f"mysql+aiomysql://127.0.0.1:{port}")
    try:
        yield engine
    finally:
        await engine.dispose()


async def query(
    conn: MySQLConnection,
    sql: str,
    cursor_class: Type[MySQLCursor] = MySQLCursorDict,
    params: Sequence[Any] = None,
    query_attributes: Dict[str, str] = None,
) -> Sequence[Any]:
    cursor = await to_thread(conn.cursor, cursor_class=cursor_class)
    if query_attributes:
        for key, value in query_attributes.items():
            cursor.add_attribute(key, value)
    await to_thread(cursor.execute, sql, *(p for p in [params] if p))
    result = await to_thread(cursor.fetchall)
    await to_thread(cursor.close)
    return result



from __future__ import annotations

import io
import struct
from dataclasses import dataclass
from datetime import datetime, date, timedelta
from typing import Iterable, Sequence, Optional, Callable, Any, Union, Tuple, Dict

from mysql_mimic.errors import MysqlError
from mysql_mimic.types import ColumnType, str_len, uint_1, uint_2, uint_4
from mysql_mimic.charset import CharacterSet


Encoder = Callable[[Any, "ResultColumn"], bytes]


class ResultColumn:
    """
    Column data for a result set

    Args:
        name: column name
        type: column type
        character_set: column character set. Only relevant for string columns.
        text_encoder: Optionally override the function used to encode values for MySQL's text protocol
        binary_encoder: Optionally override the function used to encode values for MySQL's binary protocol
    """

    def __init__(
        self,
        name: str,
        type: ColumnType,  # pylint: disable=redefined-builtin
        character_set: CharacterSet = CharacterSet.utf8mb4,
        text_encoder: Optional[Encoder] = None,
        binary_encoder: Optional[Encoder] = None,
    ):
        self.name = name
        self.type = type
        self.character_set = character_set
        self.text_encoder = text_encoder or _TEXT_ENCODERS.get(type) or _unsupported
        self.binary_encoder = (
            binary_encoder or _BINARY_ENCODERS.get(type) or _unsupported
        )

    def text_encode(self, val: Any) -> bytes:
        return self.text_encoder(self, val)

    def binary_encode(self, val: Any) -> bytes:
        return self.binary_encoder(self, val)


@dataclass
class ResultSet:
    rows: Iterable[Sequence]
    columns: Sequence[ResultColumn]

    def __bool__(self) -> bool:
        return bool(self.columns)


@dataclass
class Column:
    """
    Column data for a SHOW COLUMNS statement

    https://dev.mysql.com/doc/refman/8.0/en/show-columns.html
    """

    name: str
    type: str
    collation: str = "NULL"
    null: str = "YES"
    key: Optional[str] = None
    default: Optional[str] = None
    extra: Optional[str] = None
    privileges: Optional[str] = None
    comment: Optional[str] = None


AllowedColumn = Union[ResultColumn, str]
AllowedResult = Union[
    ResultSet, Tuple[Sequence[Sequence[Any]], Sequence[AllowedColumn]], None
]


def ensure_result_set(result: AllowedResult) -> ResultSet:
    if result is None:
        return ResultSet([], [])
    if isinstance(result, ResultSet):
        return result
    if isinstance(result, tuple):
        if len(result) != 2:
            raise MysqlError(
                f"Result tuple should be of size 2. Received: {len(result)}"
            )
        rows = result[0]
        columns = result[1]

        return ResultSet(
            rows=rows,
            columns=[_ensure_result_col(col, i, rows) for i, col in enumerate(columns)],
        )

    raise MysqlError(f"Unexpected result set type: {type(result)}")


def _ensure_result_col(
    column: AllowedColumn, idx: int, rows: Sequence[Sequence[Any]]
) -> ResultColumn:
    if isinstance(column, ResultColumn):
        return column

    if isinstance(column, str):
        value = _find_first_non_null_value(idx, rows)
        type_ = infer_type(value)
        return ResultColumn(
            name=column,
            type=type_,
        )

    raise MysqlError(f"Unexpected result column value: {column}")


def _find_first_non_null_value(
    idx: int, rows: Sequence[Sequence[Any]]
) -> Optional[Any]:
    for row in rows:
        value = row[idx]
        if value is not None:
            return value
    return None


def _binary_encode_tiny(col: ResultColumn, val: Any) -> bytes:
    return uint_1(int(bool(val)))


def _binary_encode_str(col: ResultColumn, val: Any) -> bytes:
    if not isinstance(val, bytes):
        val = str(val)

    if isinstance(val, str):
        val = val.encode(col.character_set.codec)

    return str_len(val)


def _binary_encode_date(col: ResultColumn, val: Any) -> bytes:
    if isinstance(val, (float, int)):
        val = datetime.fromtimestamp(val)

    year = val.year
    month = val.month
    day = val.day

    if isinstance(val, datetime):
        hour = val.hour
        minute = val.minute
        second = val.second
        microsecond = val.microsecond
    else:
        hour = minute = second = microsecond = 0

    if microsecond == 0:
        if hour == minute == second == 0:
            if year == month == day == 0:
                return uint_1(0)
            return b"".join([uint_1(4), uint_2(year), uint_1(month), uint_1(day)])
        return b"".join(
            [
                uint_1(7),
                uint_2(year),
                uint_1(month),
                uint_1(day),
                uint_1(hour),
                uint_1(minute),
                uint_1(second),
            ]
        )
    return b"".join(
        [
            uint_1(11),
            uint_2(year),
            uint_1(month),
            uint_1(day),
            uint_1(hour),
            uint_1(minute),
            uint_1(second),
            uint_4(microsecond),
        ]
    )


def _binary_encode_short(col: ResultColumn, val: Any) -> bytes:
    return struct.pack("<h", val)


def _binary_encode_int(col: ResultColumn, val: Any) -> bytes:
    return struct.pack("<i", val)


def _binary_encode_long(col: ResultColumn, val: Any) -> bytes:
    return struct.pack("<l", val)


def _binary_encode_longlong(col: ResultColumn, val: Any) -> bytes:
    return struct.pack("<q", val)


def _binary_encode_float(col: ResultColumn, val: Any) -> bytes:
    return struct.pack("<f", val)


def _binary_encode_double(col: ResultColumn, val: Any) -> bytes:
    return struct.pack("<d", val)


def _binary_encode_timedelta(col: ResultColumn, val: Any) -> bytes:
    days = abs(val.days)
    hours, remainder = divmod(abs(val.seconds), 3600)
    minutes, seconds = divmod(remainder, 60)
    microseconds = val.microseconds
    is_negative = val.total_seconds() < 0

    if microseconds == 0:
        if days == hours == minutes == seconds == 0:
            return uint_1(0)
        return b"".join(
            [
                uint_1(8),
                uint_1(is_negative),
                uint_4(days),
                uint_1(hours),
                uint_1(minutes),
                uint_1(seconds),
            ]
        )
    return b"".join(
        [
            uint_1(12),
            uint_1(is_negative),
            uint_4(days),
            uint_1(hours),
            uint_1(minutes),
            uint_1(seconds),
            uint_4(microseconds),
        ]
    )


def _text_encode_str(col: ResultColumn, val: Any) -> bytes:
    if not isinstance(val, bytes):
        val = str(val)

    if isinstance(val, str):
        val = val.encode(col.character_set.codec)

    return val


def _text_encode_tiny(col: ResultColumn, val: Any) -> bytes:
    return _text_encode_str(col, int(val))


def _unsupported(col: ResultColumn, val: Any) -> bytes:
    raise MysqlError(f"Unsupported column type: {col.type}")


def _noop(col: ResultColumn, val: Any) -> bytes:
    return val


# Order matters
# bool is a subclass of int
# datetime is a subclass of date
_PY_TO_MYSQL_TYPE = {
    bool: ColumnType.TINY,
    datetime: ColumnType.DATETIME,
    str: ColumnType.STRING,
    bytes: ColumnType.BLOB,
    int: ColumnType.LONGLONG,
    float: ColumnType.DOUBLE,
    date: ColumnType.DATE,
    timedelta: ColumnType.TIME,
}


_TEXT_ENCODERS: Dict[ColumnType, Encoder] = {
    ColumnType.DECIMAL: _text_encode_str,
    ColumnType.TINY: _text_encode_tiny,
    ColumnType.SHORT: _text_encode_str,
    ColumnType.LONG: _text_encode_str,
    ColumnType.FLOAT: _text_encode_str,
    ColumnType.DOUBLE: _text_encode_str,
    ColumnType.NULL: _unsupported,
    ColumnType.TIMESTAMP: _text_encode_str,
    ColumnType.LONGLONG: _text_encode_str,
    ColumnType.INT24: _text_encode_str,
    ColumnType.DATE: _text_encode_str,
    ColumnType.TIME: _text_encode_str,
    ColumnType.DATETIME: _text_encode_str,
    ColumnType.YEAR: _text_encode_str,
    ColumnType.NEWDATE: _text_encode_str,
    ColumnType.VARCHAR: _text_encode_str,
    ColumnType.BIT: _text_encode_str,
    ColumnType.TIMESTAMP2: _unsupported,
    ColumnType.DATETIME2: _unsupported,
    ColumnType.TIME2: _unsupported,
    ColumnType.TYPED_ARRAY: _unsupported,
    ColumnType.INVALID: _unsupported,
    ColumnType.BOOL: _binary_encode_tiny,
    ColumnType.JSON: _text_encode_str,
    ColumnType.NEWDECIMAL: _text_encode_str,
    ColumnType.ENUM: _text_encode_str,
    ColumnType.SET: _text_encode_str,
    ColumnType.TINY_BLOB: _text_encode_str,
    ColumnType.MEDIUM_BLOB: _text_encode_str,
    ColumnType.LONG_BLOB: _text_encode_str,
    ColumnType.BLOB: _text_encode_str,
    ColumnType.VAR_STRING: _text_encode_str,
    ColumnType.STRING: _text_encode_str,
    ColumnType.GEOMETRY: _text_encode_str,
}

_BINARY_ENCODERS: Dict[ColumnType, Encoder] = {
    ColumnType.DECIMAL: _binary_encode_str,
    ColumnType.TINY: _binary_encode_tiny,
    ColumnType.SHORT: _binary_encode_short,
    ColumnType.LONG: _binary_encode_long,
    ColumnType.FLOAT: _binary_encode_float,
    ColumnType.DOUBLE: _binary_encode_double,
    ColumnType.NULL: _unsupported,
    ColumnType.TIMESTAMP: _binary_encode_date,
    ColumnType.LONGLONG: _binary_encode_longlong,
    ColumnType.INT24: _binary_encode_long,
    ColumnType.DATE: _binary_encode_date,
    ColumnType.TIME: _binary_encode_timedelta,
    ColumnType.DATETIME: _binary_encode_date,
    ColumnType.YEAR: _binary_encode_short,
    ColumnType.NEWDATE: _unsupported,
    ColumnType.VARCHAR: _binary_encode_str,
    ColumnType.BIT: _binary_encode_str,
    ColumnType.TIMESTAMP2: _unsupported,
    ColumnType.DATETIME2: _unsupported,
    ColumnType.TIME2: _unsupported,
    ColumnType.TYPED_ARRAY: _unsupported,
    ColumnType.INVALID: _unsupported,
    ColumnType.BOOL: _binary_encode_tiny,
    ColumnType.JSON: _binary_encode_str,
    ColumnType.NEWDECIMAL: _binary_encode_str,
    ColumnType.ENUM: _binary_encode_str,
    ColumnType.SET: _binary_encode_str,
    ColumnType.TINY_BLOB: _binary_encode_str,
    ColumnType.MEDIUM_BLOB: _binary_encode_str,
    ColumnType.LONG_BLOB: _binary_encode_str,
    ColumnType.BLOB: _binary_encode_str,
    ColumnType.VAR_STRING: _binary_encode_str,
    ColumnType.STRING: _binary_encode_str,
    ColumnType.GEOMETRY: _binary_encode_str,
}


def infer_type(val: Any) -> ColumnType:
    for py_type, my_type in _PY_TO_MYSQL_TYPE.items():
        if isinstance(val, py_type):
            return my_type
    return ColumnType.VARCHAR


class NullBitmap:
    """See https://dev.mysql.com/doc/internals/en/null-bitmap.html"""

    __slots__ = ("offset", "bitmap")

    def __init__(self, bitmap: bytearray, offset: int = 0):
        self.offset = offset
        self.bitmap = bitmap

    @classmethod
    def new(cls, num_bits: int, offset: int = 0) -> NullBitmap:
        bitmap = bytearray(cls._num_bytes(num_bits, offset))
        return cls(bitmap, offset)

    @classmethod
    def from_buffer(
        cls, buffer: io.BytesIO, num_bits: int, offset: int = 0
    ) -> NullBitmap:
        bitmap = bytearray(buffer.read(cls._num_bytes(num_bits, offset)))
        return cls(bitmap, offset)

    @classmethod
    def _num_bytes(cls, num_bits: int, offset: int) -> int:
        return (num_bits + 7 + offset) // 8

    def flip(self, i: int) -> None:
        byte_position, bit_position = self._pos(i)
        self.bitmap[byte_position] |= 1 << bit_position

    def is_flipped(self, i: int) -> bool:
        byte_position, bit_position = self._pos(i)
        return bool(self.bitmap[byte_position] & (1 << bit_position))

    def _pos(self, i: int) -> Tuple[int, int]:
        byte_position = (i + self.offset) // 8
        bit_position = (i + self.offset) % 8
        return byte_position, bit_position

    def __bytes__(self) -> bytes:
        return bytes(self.bitmap)

    def __repr__(self) -> str:
        return "".join(format(b, "08b") for b in self.bitmap)



import sys
from collections.abc import Iterator


class seq(Iterator):
    """Auto-incrementing sequence with an optional maximum size"""

    def __init__(self, size: int = None):
        self.size = size
        self.value = 0

    def __next__(self) -> int:
        value = self.value
        self.value = self.value + 1
        if self.size:
            self.value = self.value % self.size
        return value

    def reset(self) -> None:
        self.value = 0


def xor(a: bytes, b: bytes) -> bytes:
    # Fast XOR implementation, according to https://stackoverflow.com/questions/29408173/byte-operations-xor-in-python
    a, b = a[: len(b)], b[: len(a)]
    int_b = int.from_bytes(b, sys.byteorder)
    int_a = int.from_bytes(a, sys.byteorder)
    int_enc = int_b ^ int_a
    return int_enc.to_bytes(len(b), sys.byteorder)



import asyncio
import random
from ssl import SSLContext
from socket import socket
from typing import Callable, Any, Dict, Optional, Sequence

from mysql_mimic.auth import IdentityProvider, SimpleIdentityProvider
from mysql_mimic.connection import Connection
from mysql_mimic.session import Session
from mysql_mimic.constants import DEFAULT_SERVER_CAPABILITIES
from mysql_mimic.stream import MysqlStream
from mysql_mimic.types import Capabilities
from mysql_mimic.utils import seq


class MaxConnectionsExceeded(Exception):
    pass


class MysqlServer:
    """
    MySQL mimic server.

    Args:
        session_factory: Callable that takes no arguments and returns a session
        capabilities: server capability flags
        server_id: set a unique server ID. This is used to generate globally unique
            connection IDs. This should be an integer between 0 and 65535.
            If left as None, a random server ID will be generated.
        identity_provider: Authentication plugins to register. Defaults to `SimpleIdentityProvider`,
            which just blindly accepts whatever `username` is given by the client.
        ssl: SSLContext instance if this server should enable TLS over connections

        **kwargs: extra keyword args passed to the asyncio start server command
    """

    _CONNECTION_ID_BITS = 16
    _MAX_CONNECTION_SEQ = 2**_CONNECTION_ID_BITS
    _MAX_SERVER_ID = 2**16

    def __init__(
        self,
        session_factory: Callable[[], Session] = Session,
        capabilities: Capabilities = DEFAULT_SERVER_CAPABILITIES,
        server_id: int = None,
        identity_provider: IdentityProvider = None,
        ssl: SSLContext = None,
        **serve_kwargs: Any,
    ):
        self.session_factory = session_factory
        self.capabilities = capabilities
        self.server_id = server_id or self._get_server_id()
        self.identity_provider = identity_provider or SimpleIdentityProvider()
        self.ssl = ssl

        self._connection_seq = seq(self._MAX_CONNECTION_SEQ)
        self._connections: Dict[int, Connection] = {}
        self._serve_kwargs = serve_kwargs
        self._server: Optional[asyncio.base_events.Server] = None

    async def _client_connected_cb(
        self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        connection_id = self._get_connection_id()
        connection = Connection(
            stream=MysqlStream(reader, writer),
            session=self.session_factory(),
            server_capabilities=self.capabilities,
            connection_id=connection_id,
            identity_provider=self.identity_provider,
            ssl=self.ssl,
        )
        self._connections[connection_id] = connection
        try:
            return await connection.start()
        finally:
            self._connections.pop(connection_id, None)

    def _get_server_id(self) -> int:
        return random.randint(0, self._MAX_SERVER_ID - 1)

    def _get_connection_id(self) -> int:
        """
        Generate a connection ID.

        MySQL connection IDs are 4 bytes.

        This is tricky for us, as there may be multiple MySQL-Mimic server instances.

        We use a concatenation of the server ID (first two bytes) and connection
        sequence (second two bytes):

        |<---server ID--->|<-conn sequence->|
         00000000 00000000 00000000 00000000

        If incremental connection IDs aren't manually provided, this isn't guaranteed
        to be unique. But collisions should be highly unlikely.
        """
        if len(self._connections) >= self._MAX_CONNECTION_SEQ:
            raise MaxConnectionsExceeded()
        server_id_prefix = (
            self.server_id % self._MAX_SERVER_ID
        ) << self._CONNECTION_ID_BITS

        connection_id = server_id_prefix + next(self._connection_seq)

        # Avoid connection ID collisions in the unlikely chance that a connection is
        # alive longer than it takes for the sequence to reset.
        while connection_id in self._connections:
            connection_id = server_id_prefix + next(self._connection_seq)

        return connection_id

    async def start_server(self, **kwargs: Any) -> None:
        """
        Start an asyncio socket server.

        Args:
            **kwargs: keyword args passed to `asyncio.start_server`
        """
        kw = {}
        kw.update(self._serve_kwargs)
        kw.update(kwargs)
        if "port" not in kw:
            kw["port"] = 3306
        self._server = await asyncio.start_server(self._client_connected_cb, **kw)

    async def start_unix_server(self, **kwargs: Any) -> None:
        """
        Start an asyncio unix socket server.

        Args:
            **kwargs: keyword args passed to `asyncio.start_unix_server`
        """
        kw = {}
        kw.update(self._serve_kwargs)
        kw.update(kwargs)
        self._server = await asyncio.start_unix_server(self._client_connected_cb, **kw)

    async def serve_forever(self, **kwargs: Any) -> None:
        """
        Start accepting connections until the coroutine is cancelled.

        If a server isn't running, this starts a server with `start_server`.

        Args:
            **kwargs: keyword args passed to `start_server`
        """
        if not self._server:
            await self.start_server(**kwargs)
        assert self._server is not None
        await self._server.serve_forever()

    def close(self) -> None:
        """
        Stop serving.

        The server is closed asynchronously -
        use the `wait_closed` coroutine to wait until the server is closed.
        """
        if self._server:
            self._server.close()

    async def wait_closed(self) -> None:
        """Wait until the `close` method completes."""
        if self._server:
            await self._server.wait_closed()

    def sockets(self) -> Sequence[socket]:
        """Get sockets the server is listening on."""
        if self._server:
            return self._server.sockets
        return ()



from enum import IntEnum


class ErrorCode(IntEnum):
    """https://dev.mysql.com/doc/mysql-errors/8.0/en/server-error-reference.html"""

    CON_COUNT_ERROR = 1040
    HANDSHAKE_ERROR = 1043
    ACCESS_DENIED_ERROR = 1045
    NO_DB_ERROR = 1046
    PARSE_ERROR = 1064
    EMPTY_QUERY = 1065
    UNKNOWN_PROCEDURE = 1106
    UNKNOWN_SYSTEM_VARIABLE = 1193
    UNKNOWN_COM_ERROR = 1047
    UNKNOWN_ERROR = 1105
    WRONG_VALUE_FOR_VAR = 1231
    NOT_SUPPORTED_YET = 1235
    MALFORMED_PACKET = 1835
    USER_DOES_NOT_EXIST = 3162
    PLUGIN_REQUIRES_REGISTRATION = 4055


# For more info, see https://dev.mysql.com/doc/refman/8.0/en/error-message-elements.html
SQLSTATES = {
    ErrorCode.CON_COUNT_ERROR: b"08004",
    ErrorCode.ACCESS_DENIED_ERROR: b"28000",
    ErrorCode.HANDSHAKE_ERROR: b"08S01",
    ErrorCode.NO_DB_ERROR: b"3D000",
    ErrorCode.PARSE_ERROR: b"42000",
    ErrorCode.EMPTY_QUERY: b"42000",
    ErrorCode.UNKNOWN_PROCEDURE: b"42000",
    ErrorCode.UNKNOWN_COM_ERROR: b"08S01",
    ErrorCode.WRONG_VALUE_FOR_VAR: b"42000",
    ErrorCode.NOT_SUPPORTED_YET: b"42000",
}


def get_sqlstate(code: ErrorCode) -> bytes:
    return SQLSTATES.get(code, b"HY000")


class MysqlError(Exception):
    def __init__(self, msg: str, code: ErrorCode = ErrorCode.UNKNOWN_ERROR):
        super().__init__(f"{code}: {msg}")
        self.msg = msg
        self.code = code



from mysql_mimic.types import Capabilities

DEFAULT_SERVER_CAPABILITIES = (
    Capabilities.CLIENT_PROTOCOL_41
    | Capabilities.CLIENT_DEPRECATE_EOF
    | Capabilities.CLIENT_CONNECT_WITH_DB
    | Capabilities.CLIENT_QUERY_ATTRIBUTES
    | Capabilities.CLIENT_CONNECT_ATTRS
    | Capabilities.CLIENT_PLUGIN_AUTH
    | Capabilities.CLIENT_PLUGIN_AUTH_LENENC_CLIENT_DATA
    | Capabilities.CLIENT_SECURE_CONNECTION
    | Capabilities.CLIENT_LONG_PASSWORD
)



import re
from typing import Optional, Any, Sequence, List

from mysql_mimic.session import Session
from mysql_mimic.charset import CharacterSet
from mysql_mimic.errors import MysqlError, ErrorCode
from mysql_mimic.results import ResultColumn, ResultSet, Column
from mysql_mimic.types import ColumnType
from mysql_mimic.variables import SystemVariables


class Admin:
    """
    Administration Statements (plus some other things)
    https://dev.mysql.com/doc/refman/8.0/en/sql-server-administration-statements.html
    """

    # Regex for finding "information functions" to replace in SQL statements
    REGEX_INFO_FUNC = re.compile(
        r"""
            (
              ((?P<func>
                (CONNECTION_ID)
                |(USER)
                |(CURRENT_USER)
                |(SYSTEM_USER)
                |(SESSION_USER)
                |(VERSION)
                |(DATABASE)
              )\(\))
            |
              (?P<current_user>CURRENT_USER)
            )
            (?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)
        """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for finding "session variables" to replace in SQL statements
    REGEX_SESSION_VAR = re.compile(
        r"""
            @@(?P<var>\w+)
            (?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)
        """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for determining if a statement is an "admin statement"
    REGEX_CMD = re.compile(
        r"^\s*(?P<cmd>(SET\s)|(SHOW\s)|(ROLLBACK))(?P<this>.*)$",
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SET VARIABLES" statement
    # Not yet supported:
    # 1. Setting variable to result of scalar subquery
    # 2. Setting variable to another variable
    # 3. Setting multiple variables at once
    REGEX_SET_VAR = re.compile(
        r"""
        ^\s*
        (
            (?P<global>(GLOBAL\s+)|(@@GLOBAL\.))
            |(?P<persist>(PERSIST\s+)|(@@PERSIST\.))
            |(?P<persist_only>(PERSIST_ONLY\s+)|(@@PERSIST_ONLY\.))
            |(?P<session>(SESSION\s+)|(@@SESSION\.)|(@@))
            |(?P<user>@)
        )?
        (?P<var_name>(\w+))
        \s*
        =
        \s*(?P<value>.*?)
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SET NAMES" statement
    REGEX_SET_NAMES = re.compile(
        r"""
        ^\s*
        NAMES\s+
        '?(?P<charset_name>\w+)'?
        \s*
        (
            (COLLATE\s+'?(?P<collation_name>\w+)'?)
            |
            (DEFAULT)
        )?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW VARIABLES" statement
    REGEX_SHOW_VARS = re.compile(
        r"""
        ^\s*
        (
            (?P<global>GLOBAL)
            |(?P<session>SESSION)
            \s*
        )?
        VARIABLES\s*
        (LIKE\s*'(?P<like>[\w%]+)'\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW COLUMNS" statement
    REGEX_SHOW_COLUMNS = re.compile(
        r"""
        ^\s*
        ((?P<extended>EXTENDED)\s*)?
        ((?P<full>FULL)\s*)?
        ((COLUMNS)|(FIELDS))\s*
        ((FROM)|(IN))\s*(`?(?P<impl_db_name>\w+)`?\.)?`?(?P<tbl_name>\w+)`?\s*
        (((FROM)|(IN))\s*`?(?P<expl_db_name>\w+)`?\s*)?
        (LIKE\s*'(?P<like>[\w%]+)'\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW INDEX" statement
    REGEX_SHOW_INDEX = re.compile(
        r"""
        ^\s*
        ((?P<extended>EXTENDED)\s*)?
        ((?P<full>FULL)\s*)?
        ((INDEX)|(INDEXES)|(KEYS))\s*
        ((FROM)|(IN))\s*(`?(?P<impl_db_name>\w+)`?\.)?`?(?P<tbl_name>\w+)`?\s*
        (((FROM)|(IN))\s*`?(?P<expl_db_name>\w+)`?\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW TABLES" statement
    REGEX_SHOW_TABLES = re.compile(
        r"""
        ^\s*
        ((?P<extended>EXTENDED)\s*)?
        ((?P<full>FULL)\s*)?
        TABLES\s*
        (((FROM)|(IN))\s*`?(?P<db_name>\w+)`?\s*)?
        (LIKE\s*'(?P<like>[\w%]+)'\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    # Regex for parsing a "SHOW DATABASE" statement
    REGEX_SHOW_DATABASES = re.compile(
        r"""
        ^\s*
        ((DATABASES)|(SCHEMAS))\s*
        (LIKE\s*'(?P<like>[\w%]+)'\s*)?
        $
    """,
        re.IGNORECASE | re.VERBOSE,
    )

    def __init__(
        self, connection_id: int, session: Session, variables: SystemVariables
    ):
        self.connection_id = connection_id
        self.session = session
        self.database: Optional[str] = None
        self.username: Optional[str] = None
        self.vars = variables

    def replace_variables(self, sql: str) -> str:
        sql = self.REGEX_INFO_FUNC.sub(self._replace_info_func, sql)
        return self.REGEX_SESSION_VAR.sub(self._replace_session_var, sql)

    def _replace_info_func(self, matchobj: re.Match) -> str:
        func = (matchobj.group("func") or matchobj.group("current_user")).lower()
        if func == "connection_id":
            return str(self.connection_id)
        if func in {"user", "session_user", "system_user"}:
            return f"'{self.vars.external_user}'" if self.vars.external_user else "NULL"
        if func in {"current_user"}:
            return f"'{self.username}'" if self.username else "NULL"
        if func == "version":
            return f"'{self.vars.mysql_version}'"
        if func == "database":
            return f"'{self.database}'" if self.database else "NULL"
        raise MysqlError(f"Failed to parse system information function: {func}")

    def _replace_session_var(self, matchobj: re.Match) -> str:
        var = matchobj.group("var").lower()
        if var in self.vars:
            return f"'{self.vars[var]}'"
        raise MysqlError(f"Unknown variable: {var}", ErrorCode.UNKNOWN_SYSTEM_VARIABLE)

    async def parse(self, sql: str) -> Optional[ResultSet]:
        m = self.REGEX_CMD.match(sql)
        if not m:
            return None

        cmd = m.group("cmd").strip().lower()
        this = m.group("this").strip().lower()

        if cmd == "set":
            return await self._parse_set(this)
        if cmd == "show":
            return await self._parse_show(this)
        if cmd == "rollback":
            await self.session.rollback()
            return ResultSet([], [])
        raise MysqlError("Failed to parse command", ErrorCode.PARSE_ERROR)

    async def _parse_set(self, this: str) -> Optional[ResultSet]:
        m = self.REGEX_SET_NAMES.match(this)
        if m:
            return await self._parse_set_names(
                charset_name=m.group("charset_name"),
                collation_name=m.group("collation_name"),
            )

        m = self.REGEX_SET_VAR.match(this)
        if m:
            return await self._parse_set_variables(
                global_=m.group("global"),
                persist=m.group("persist"),
                persist_only=m.group("persist_only"),
                session=m.group("session"),
                user=m.group("user"),
                var_name=m.group("var_name"),
                value=m.group("value"),
            )

        raise MysqlError("Unsupported SET command", ErrorCode.NOT_SUPPORTED_YET)

    async def _set_variable(self, key: str, val: Any) -> None:
        self.vars[key] = val
        await self.session.set(**{key: val})

    async def _parse_set_names(
        self, charset_name: str, collation_name: str
    ) -> Optional[ResultSet]:
        await self._set_variable("character_set_client", charset_name)
        await self._set_variable("character_set_connection", charset_name)
        await self._set_variable("character_set_results", charset_name)
        await self._set_variable(
            "collation_connection",
            (collation_name or CharacterSet[charset_name].default_collation.name),
        )
        return ResultSet([], [])

    async def _parse_set_variables(
        self,
        global_: Optional[str],
        persist: Optional[str],
        persist_only: Optional[str],
        session: Optional[str],
        user: Optional[str],
        var_name: Optional[str],
        value: Optional[str],
    ) -> Optional[ResultSet]:  # pylint: disable=unused-argument
        if global_ or persist or persist_only or user:
            raise MysqlError(
                "Only setting session variables is supported",
                ErrorCode.NOT_SUPPORTED_YET,
            )

        assert value is not None
        assert var_name is not None
        result: Any

        value_lower = value.lower()
        if value_lower in {"true", "on"}:
            result = True
        elif value_lower in {"false", "off"}:
            result = False
        elif value_lower in {"default", "null"}:
            result = self.vars.defaults.get(var_name)
        elif value[0] == value[-1] == "'":
            result = value[1:-1]
        else:
            try:
                result = int(value)
            except ValueError:
                try:
                    result = float(value)
                except ValueError as e:
                    raise MysqlError(
                        f"Unexpected variable value: {value}",
                        ErrorCode.WRONG_VALUE_FOR_VAR,
                    ) from e

        await self._set_variable(var_name, result)
        return ResultSet([], [])

    async def _parse_show(self, this: str) -> Optional[ResultSet]:
        m = self.REGEX_SHOW_VARS.match(this)
        if m:
            return await self._parse_show_variables(
                global_=bool(m.group("global")),
                session=bool(m.group("session")),
                like=self._like_to_regex(m.group("like")),
            )
        m = self.REGEX_SHOW_COLUMNS.match(this)
        if m:
            return await self._parse_show_columns(
                extended=bool(m.group("extended")),
                full=bool(m.group("full")),
                db_name=m.group("expl_db_name") or m.group("impl_db_name"),
                tbl_name=m.group("tbl_name"),
                like=self._like_to_regex(m.group("like")),
            )
        m = self.REGEX_SHOW_INDEX.match(this)
        if m:
            return await self._parse_show_index(
                extended=bool(m.group("extended")),
                db_name=m.group("expl_db_name") or m.group("impl_db_name"),
                tbl_name=m.group("tbl_name"),
            )
        m = self.REGEX_SHOW_TABLES.match(this)
        if m:
            return await self._parse_show_tables(
                extended=bool(m.group("extended")),
                full=bool(m.group("full")),
                db_name=m.group("db_name"),
                like=self._like_to_regex(m.group("like")),
            )
        m = self.REGEX_SHOW_DATABASES.match(this)
        if m:
            return await self._parse_show_databases(
                like=self._like_to_regex(m.group("like")),
            )

        raise MysqlError(
            f"Unsupported SHOW command: {this}", ErrorCode.NOT_SUPPORTED_YET
        )

    def _like_to_regex(self, like: Optional[str]) -> re.Pattern:
        if like is None:
            return re.compile(r".*")
        like = like.replace("%", ".*")
        like = like.replace("_", ".")
        return re.compile(like)

    async def _parse_show_columns(
        self,
        extended: bool,
        full: bool,
        db_name: Optional[str],
        tbl_name: Optional[str],
        like: re.Pattern,
    ) -> Optional[ResultSet]:  # pylint: disable=unused-argument
        db_name = db_name or self.database
        if not db_name:
            raise MysqlError("No database selected", ErrorCode.NO_DB_ERROR)

        columns = [
            c if isinstance(c, Column) else Column(**c)
            for c in await self.session.show_columns(db_name, tbl_name or "")
        ]
        columns = [c for c in columns if like.match(c.name)]

        result_columns = [
            ResultColumn("Field", ColumnType.STRING),
            ResultColumn("Type", ColumnType.STRING),
            ResultColumn("Null", ColumnType.STRING),
            ResultColumn("Key", ColumnType.STRING),
            ResultColumn("Default", ColumnType.STRING),
            ResultColumn("Extra", ColumnType.STRING),
        ]
        if full:
            result_columns.extend(
                [
                    ResultColumn("Collation", ColumnType.STRING),
                    ResultColumn("Privileges", ColumnType.STRING),
                    ResultColumn("Comment", ColumnType.STRING),
                ]
            )
        rows = []
        for c in columns:
            row = [c.name, c.type, c.null, c.key, c.default, c.extra]
            if full:
                row.extend(
                    [
                        c.collation,
                        c.privileges,
                        c.comment,
                    ]
                )
            rows.append(row)
        return ResultSet(rows=rows, columns=result_columns)

    async def _parse_show_index(
        self, extended: bool, db_name: Optional[str], tbl_name: Optional[str]
    ) -> ResultSet:  # pylint: disable=unused-argument
        result_columns = [
            ResultColumn("Table", ColumnType.STRING),
            ResultColumn("Non_unique", ColumnType.TINY),
            ResultColumn("Key_name", ColumnType.STRING),
            ResultColumn("Seq_in_index", ColumnType.LONGLONG),
            ResultColumn("Column_name", ColumnType.STRING),
            ResultColumn("Collation", ColumnType.STRING),
            ResultColumn("Cardinality", ColumnType.LONGLONG),
            ResultColumn("Sub_part", ColumnType.STRING),
            ResultColumn("Packed", ColumnType.STRING),
            ResultColumn("Null", ColumnType.STRING),
            ResultColumn("Index_type", ColumnType.STRING),
            ResultColumn("Comment", ColumnType.STRING),
            ResultColumn("Index_comment", ColumnType.STRING),
            ResultColumn("Visible", ColumnType.STRING),
            ResultColumn("Expression", ColumnType.STRING),
        ]
        return ResultSet(rows=[], columns=result_columns)

    async def _parse_show_tables(
        self, extended: bool, full: bool, db_name: Optional[str], like: re.Pattern
    ) -> ResultSet:  # pylint: disable=unused-argument
        db_name = db_name or self.database

        if not db_name:
            raise MysqlError("No database selected", ErrorCode.NO_DB_ERROR)

        rows: List[Sequence[str]] = [
            (t,) for t in await self.session.show_tables(db_name) if like.match(t)
        ]
        columns = [ResultColumn(f"Tables_in_{db_name}", ColumnType.STRING)]

        if full:
            rows = [(r[0], "BASE TABLE") for r in rows]
            columns.append(ResultColumn("Table_type", ColumnType.STRING))

        return ResultSet(rows=rows, columns=columns)

    async def _parse_show_databases(self, like: re.Pattern) -> ResultSet:
        databases = await self.session.show_databases()
        rows = [(db,) for db in databases if like.match(db)]
        return ResultSet(
            rows=rows, columns=[ResultColumn("Database", ColumnType.STRING)]
        )

    async def _parse_show_variables(
        self, global_: bool, session: bool, like: re.Pattern
    ) -> ResultSet:  # pylint: disable=unused-argument
        rows = list(self.vars.items())
        rows = [(k, v) for k, v in rows if like.match(k)]
        return ResultSet(
            rows=rows,
            columns=[
                ResultColumn("Variable_name", ColumnType.STRING),
                ResultColumn("Value", ColumnType.STRING),
            ],
        )



import re
from dataclasses import dataclass
from typing import Optional, Dict, Iterable


# Borrowed from mysql-connector-python
REGEX_PARAM = re.compile(r"""\?(?=(?:[^"'`]*["'`][^"'`]*["'`])*[^"'`]*$)""")


@dataclass
class PreparedStatement:
    stmt_id: int
    sql: str
    num_params: int
    param_buffers: Optional[Dict[int, bytearray]] = None
    cursor: Optional[Iterable] = None



import io
import struct

import pytest

from mysql_mimic import types


def test_column_definition() -> None:
    assert types.ColumnDefinition.NOT_NULL_FLAG == 1
    assert types.ColumnDefinition.PRI_KEY_FLAG == 2
    assert types.ColumnDefinition.FIELD_IS_INVISIBLE == 1 << 30
    assert (
        types.ColumnDefinition.NOT_NULL_FLAG | types.ColumnDefinition.PRI_KEY_FLAG == 3
    )


def test_capabilities() -> None:
    assert types.Capabilities.CLIENT_LONG_PASSWORD == 1
    assert types.Capabilities.CLIENT_FOUND_ROWS == 2
    assert types.Capabilities.CLIENT_PROTOCOL_41 == 512
    assert types.Capabilities.CLIENT_DEPRECATE_EOF == 1 << 24
    assert (
        types.Capabilities.CLIENT_LONG_PASSWORD | types.Capabilities.CLIENT_FOUND_ROWS
        == 3
    )


def test_server_status() -> None:
    assert types.ServerStatus.SERVER_STATUS_IN_TRANS == 1
    assert types.ServerStatus.SERVER_STATUS_AUTOCOMMIT == 2
    assert types.ServerStatus.SERVER_STATUS_LAST_ROW_SENT == 128
    assert (
        types.ServerStatus.SERVER_STATUS_IN_TRANS
        | types.ServerStatus.SERVER_STATUS_AUTOCOMMIT
        == 3
    )


def test_int_1() -> None:
    assert types.uint_1(0) == b"\x00"
    assert types.uint_1(1) == b"\x01"
    assert types.uint_1(255) == b"\xff"
    pytest.raises(struct.error, types.uint_1, 256)


def test_int_2() -> None:
    assert types.uint_2(0) == b"\x00\x00"
    assert types.uint_2(1) == b"\x01\x00"
    assert types.uint_2(255) == b"\xff\x00"
    assert types.uint_2(2**8) == b"\x00\x01"
    assert types.uint_2(2**16 - 1) == b"\xff\xff"
    pytest.raises(struct.error, types.uint_2, 2**16)


def test_int_3() -> None:
    assert types.uint_3(0) == b"\x00\x00\x00"
    assert types.uint_3(1) == b"\x01\x00\x00"
    assert types.uint_3(255) == b"\xff\x00\x00"
    assert types.uint_3(2**8) == b"\x00\x01\x00"
    assert types.uint_3(2**16) == b"\x00\x00\x01"
    pytest.raises(struct.error, types.uint_3, 2**24)


def test_int_4() -> None:
    assert types.uint_4(0) == b"\x00\x00\x00\x00"
    assert types.uint_4(1) == b"\x01\x00\x00\x00"
    assert types.uint_4(255) == b"\xff\x00\x00\x00"
    assert types.uint_4(2**8) == b"\x00\x01\x00\x00"
    assert types.uint_4(2**16) == b"\x00\x00\x01\x00"
    assert types.uint_4(2**24) == b"\x00\x00\x00\x01"
    pytest.raises(struct.error, types.uint_4, 2**32)


def test_int_6() -> None:
    assert types.uint_6(0) == b"\x00\x00\x00\x00\x00\x00"
    assert types.uint_6(1) == b"\x01\x00\x00\x00\x00\x00"
    assert types.uint_6(255) == b"\xff\x00\x00\x00\x00\x00"
    assert types.uint_6(2**8) == b"\x00\x01\x00\x00\x00\x00"
    assert types.uint_6(2**16) == b"\x00\x00\x01\x00\x00\x00"
    assert types.uint_6(2**24) == b"\x00\x00\x00\x01\x00\x00"
    assert types.uint_6(2**32) == b"\x00\x00\x00\x00\x01\x00"
    assert types.uint_6(2**40) == b"\x00\x00\x00\x00\x00\x01"
    pytest.raises(struct.error, types.uint_6, 2**48)


def test_int_8() -> None:
    assert types.uint_8(0) == b"\x00\x00\x00\x00\x00\x00\x00\x00"
    assert types.uint_8(1) == b"\x01\x00\x00\x00\x00\x00\x00\x00"
    assert types.uint_8(255) == b"\xff\x00\x00\x00\x00\x00\x00\x00"
    assert types.uint_8(2**8) == b"\x00\x01\x00\x00\x00\x00\x00\x00"
    assert types.uint_8(2**16) == b"\x00\x00\x01\x00\x00\x00\x00\x00"
    assert types.uint_8(2**24) == b"\x00\x00\x00\x01\x00\x00\x00\x00"
    assert types.uint_8(2**32) == b"\x00\x00\x00\x00\x01\x00\x00\x00"
    assert types.uint_8(2**56) == b"\x00\x00\x00\x00\x00\x00\x00\x01"
    pytest.raises(struct.error, types.uint_8, 2**64)


def test_int_len() -> None:
    assert types.uint_len(0) == b"\x00"
    assert types.uint_len(1) == b"\x01"
    assert types.uint_len(250) == b"\xfa"
    assert types.uint_len(251) == b"\xfc\xfb\x00"
    assert types.uint_len(2**8) == b"\xfc\x00\x01"
    assert types.uint_len(2**16) == b"\xfd\x00\x00\x01"
    assert types.uint_len(2**24) == b"\xfe\x00\x00\x00\x01\x00\x00\x00\x00"
    assert types.uint_len(2**32) == b"\xfe\x00\x00\x00\x00\x01\x00\x00\x00"
    assert types.uint_len(2**56) == b"\xfe\x00\x00\x00\x00\x00\x00\x00\x01"
    pytest.raises(struct.error, types.uint_len, 2**64)


def test_str_fixed() -> None:
    assert types.str_fixed(0, b"kelsin") == b""
    assert types.str_fixed(1, b"kelsin") == b"k"
    assert types.str_fixed(6, b"kelsin") == b"kelsin"
    assert types.str_fixed(8, b"kelsin") == b"kelsin\x00\x00"


def test_str_null() -> None:
    assert types.str_null(b"") == b"\x00"
    assert types.str_null(b"kelsin") == b"kelsin\x00"


def test_str_len() -> None:
    assert types.str_len(b"") == b"\x00"
    assert types.str_len(b"kelsin") == b"\x06kelsin"

    big_str = bytes(256)
    assert types.str_len(big_str) == b"\xfc\x00\x01" + big_str


def test_str_rest() -> None:
    assert types.str_rest(b"") == b""
    assert types.str_rest(b"kelsin") == b"kelsin"

    big_str = bytes(256)
    assert types.str_rest(big_str) == big_str


def test_read_int_1() -> None:
    reader = io.BytesIO(b"\x01")
    assert types.read_uint_1(reader) == 1
    reader = io.BytesIO(b"\x00\x00")
    assert types.read_uint_1(reader) == 0
    assert types.read_uint_1(reader) == 0


def test_read_int_2() -> None:
    reader = io.BytesIO(b"\x00\x01")
    assert types.read_uint_2(reader) == 256


def test_read_int_3() -> None:
    reader = io.BytesIO(b"\x00\x00\x01")
    assert types.read_uint_3(reader) == 2**16


def test_read_int_4() -> None:
    reader = io.BytesIO(b"\x01\x00\x00\x00")
    assert types.read_uint_4(reader) == 1
    reader = io.BytesIO(b"\x00\x00\x01\x00")
    assert types.read_uint_4(reader) == 2**16
    reader = io.BytesIO(b"\x00\x00\x00\x01")
    assert types.read_uint_4(reader) == 2**24


def test_read_int_6() -> None:
    reader = io.BytesIO(b"\x01\x00\x00\x00\x00\x00")
    assert types.read_uint_6(reader) == 1
    reader = io.BytesIO(b"\x00\x00\x00\x00\x01\x00")
    assert types.read_uint_6(reader) == 2**32
    reader = io.BytesIO(b"\x00\x00\x00\x00\x00\x01")
    assert types.read_uint_6(reader) == 2**40


def test_read_int_8() -> None:
    reader = io.BytesIO(b"\x01\x00\x00\x00\x00\x00\x00\x00")
    assert types.read_uint_8(reader) == 1
    reader = io.BytesIO(b"\x00\x00\x00\x00\x00\x00\x01\x00")
    assert types.read_uint_8(reader) == 2**48
    reader = io.BytesIO(b"\x00\x00\x00\x00\x00\x00\x00\x01")
    assert types.read_uint_8(reader) == 2**56


def test_read_int_len() -> None:
    reader = io.BytesIO(b"\x01")
    assert types.read_uint_len(reader) == 1
    reader = io.BytesIO(b"\xfa")
    assert types.read_uint_len(reader) == 250
    reader = io.BytesIO(b"\xfc\x00\x01")
    assert types.read_uint_len(reader) == 256
    reader = io.BytesIO(b"\xfd\x00\x00\x01")
    assert types.read_uint_len(reader) == 2**16
    reader = io.BytesIO(b"\xfe\x00\x00\x00\x00\x00\x00\x00\x01")
    assert types.read_uint_len(reader) == 2**56


def test_read_str_fixed() -> None:
    reader = io.BytesIO(b"kelsin")
    assert types.read_str_fixed(reader, 1) == b"k"
    assert types.read_str_fixed(reader, 2) == b"el"
    assert types.read_str_fixed(reader, 3) == b"sin"


def test_read_str_null() -> None:
    reader = io.BytesIO(b"\x00")
    assert types.read_str_null(reader) == b""
    reader = io.BytesIO(b"kelsin\x00")
    assert types.read_str_null(reader) == b"kelsin"
    reader = io.BytesIO(b"kelsin\x00foo")
    assert types.read_str_null(reader) == b"kelsin"


def test_read_str_len() -> None:
    reader = io.BytesIO(b"\x01kelsin")
    assert types.read_str_len(reader) == b"k"

    big_str = bytes(256)
    reader = io.BytesIO(b"\xfc\x00\x01" + big_str)
    assert types.read_str_len(reader) == big_str


def test_read_str_rest() -> None:
    reader = io.BytesIO(b"kelsin")
    assert types.read_str_rest(reader) == b"kelsin"
    reader = io.BytesIO(b"kelsin\x00foo")
    assert types.read_str_rest(reader) == b"kelsin\x00foo"



import io
from contextlib import closing
from datetime import date, datetime, timedelta
from functools import partial
from typing import Any, Callable, Awaitable, Sequence, Dict, List, Tuple

import pytest
import pytest_asyncio
import sqlalchemy.engine
from mysql.connector import MySQLConnection
from sqlalchemy import text
import aiomysql

from mysql_mimic import ResultColumn, ResultSet, MysqlServer
from mysql_mimic.charset import CharacterSet
from mysql_mimic.results import AllowedResult
from mysql_mimic.types import ColumnType
from tests.conftest import PreparedDictCursor, query, MockSession, ConnectFixture

QueryFixture = Callable[[str], Awaitable[Sequence[Dict[str, Any]]]]


@pytest_asyncio.fixture(
    params=["mysql.connector", "mysql.connector(prepared)", "aiomysql", "sqlalchemy"]
)
async def query_fixture(
    mysql_connector_conn: MySQLConnection,
    aiomysql_conn: aiomysql.Connection,
    session: MockSession,
    sqlalchemy_engine: sqlalchemy.engine.Engine,
    request: Any,
) -> QueryFixture:
    if request.param == "mysql.connector":

        async def q1(sql: str) -> Sequence[Dict[str, Any]]:
            return await query(mysql_connector_conn, sql)

        return q1

    if request.param == "mysql.connector(prepared)":

        async def q2(sql: str) -> Sequence[Dict[str, Any]]:
            return await query(
                mysql_connector_conn, sql, cursor_class=PreparedDictCursor
            )

        return q2

    if request.param == "aiomysql":

        async def q3(sql: str) -> Sequence[Dict[str, Any]]:
            async with aiomysql_conn.cursor(aiomysql.DictCursor) as cur:
                await cur.execute(sql)
                return await cur.fetchall()

        return q3

    if request.param == "sqlalchemy":

        async def q4(sql: str) -> Sequence[Dict[str, Any]]:
            # Sqlalchemy fires off a bunch of metadata queries when connecting.
            # We'll just route things like "SELECT VERSION()" to sqlite, which should be
            # fine because the connection replaces "VERSION()" with the version name.
            session.use_sqlite = True
            async with sqlalchemy_engine.connect() as conn:
                session.use_sqlite = False
                cursor = await conn.execute(text(sql))
                return cursor.mappings().all()

        return q4

    raise Exception("Unexpected fixture param")


EXPLICIT_TYPE_TESTS = [
    (
        ResultSet(
            rows=[(input_,)],
            columns=[
                ResultColumn(
                    name="b",
                    character_set=CharacterSet.utf8mb4,
                    type=type_,
                )
            ],
        ),
        [{"b": output}],
    )
    for input_, type_, output in [
        ("♥".encode("utf-8"), ColumnType.VARCHAR, "♥"),
        ("♥".encode("utf-8"), ColumnType.BLOB, "♥"),
        (b"\xe2\x99\xa5", ColumnType.BLOB, "♥"),
        (1, ColumnType.TINY, True),
        (2, ColumnType.SHORT, 2),
        (2, ColumnType.INT24, 2),
        (2, ColumnType.LONG, 2),
        (2, ColumnType.LONGLONG, 2),
        (1.0, ColumnType.FLOAT, 1.0),
        (1.0, ColumnType.DOUBLE, 1.0),
    ]
]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "rv, expected",
    [
        (((("x",),), ("b",)), [{"b": "x"}]),
        (([["1"]], ["b"]), [{"b": "1"}]),
        (([(1,)], ("b",)), [{"b": 1}]),
        (([(1.0,)], ("b",)), [{"b": 1.0}]),
        (([(b"x",)], ("b",)), [{"b": "x"}]),
        (([(True,)], ("b",)), [{"b": True}]),
        (([(date(2021, 1, 1),)], ("b",)), [{"b": date(2021, 1, 1)}]),
        (
            ([(datetime(2021, 1, 1, 1, 1, 1),)], ("b",)),
            [{"b": datetime(2021, 1, 1, 1, 1, 1)}],
        ),
        (([(timedelta(seconds=60),)], ("b",)), [{"b": timedelta(seconds=60)}]),
        (([(None,)], ["b"]), [{"b": None}]),
        (([(None, 1, 1)], ["a", "b", "c"]), [{"a": None, "b": 1, "c": 1}]),
        (([(1, None, 1)], ["a", "b", "c"]), [{"a": 1, "b": None, "c": 1}]),
        (([(1, 1, None)], ["a", "b", "c"]), [{"a": 1, "b": 1, "c": None}]),
        (([[None], [1]], ["b"]), [{"b": None}, {"b": 1}]),
        (
            ResultSet(
                rows=[("hello",)],
                columns=[
                    ResultColumn(
                        name="b",
                        character_set=CharacterSet.utf8mb4,
                        type=ColumnType.VARCHAR,
                        text_encoder=lambda col, val: b"world",
                        binary_encoder=lambda col, val: b"\x05world",
                    )
                ],
            ),
            [{"b": "world"}],
        ),
        *EXPLICIT_TYPE_TESTS,
    ],
)
async def test_query(
    session: MockSession,
    server: MysqlServer,
    rv: AllowedResult,
    expected: List[Dict[str, Any]],
    query_fixture: QueryFixture,
) -> None:
    session.return_value = rv
    result = await query_fixture("SELECT b FROM a")
    assert expected == result


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "sql, params, expected",
    [
        ("SELECT ?", ("1",), "SELECT '1'"),
        ("SELECT '?'", (), "SELECT '?'"),
        ("SELECT ?", (1,), "SELECT 1"),
        ("SELECT ?", (0,), "SELECT 0"),
        ("SELECT ?", (-1,), "SELECT -1"),
        ("SELECT ?", (255,), "SELECT 255"),
        ("SELECT ?", (-128,), "SELECT -128"),
        ("SELECT ?", (65535,), "SELECT 65535"),
        ("SELECT ?", (-32767,), "SELECT -32767"),
        ("SELECT ?", (4294967295,), "SELECT 4294967295"),
        ("SELECT ?", (-2147483648,), "SELECT -2147483648"),
        ("SELECT ?", (18446744073709551615,), "SELECT 18446744073709551615"),
        ("SELECT ?", (-9223372036854775808,), "SELECT -9223372036854775808"),
        ("SELECT ?", (1.1,), "SELECT 1.1"),
        ("SELECT ?", (1.7e308,), "SELECT 1.7e+308"),
        ("SELECT ?", (None,), "SELECT NULL"),
        ("SELECT ?", (b"hello",), "SELECT 'hello'"),
        ("SELECT ?", (io.BytesIO(b"hello"),), "SELECT 'hello'"),
        ("SELECT ?, ?", ("1", "1"), "SELECT '1', '1'"),
        (
            "SELECT ?, ?, ?, ?",
            ("1", None, io.BytesIO(b"hello"), 1),
            "SELECT '1', NULL, 'hello', 1",
        ),
    ],
)
async def test_prepared_stmt(
    session: MockSession,
    server: MysqlServer,
    mysql_connector_conn: MySQLConnection,
    sql: str,
    params: Tuple[Any],
    expected: str,
) -> None:
    session.echo = True
    result = await query(
        conn=mysql_connector_conn,
        sql=sql,
        cursor_class=PreparedDictCursor,
        params=params,
    )
    assert expected == result[0]["sql"]


@pytest.mark.asyncio
async def test_init(port: int, session: MockSession, server: MysqlServer) -> None:
    async with aiomysql.connect(
        port=port, user="levon_helm", db="db", program_name="test"
    ):
        connection = session.connection
        assert connection is not None
        assert connection.username == "levon_helm"
        assert connection.client_connect_attrs["program_name"] == "test"
        assert connection.database == "db"
        connection.username = "robbie_robertson"
        assert connection.username == "robbie_robertson"


@pytest.mark.asyncio
async def test_connection_id(port: int, server: MysqlServer) -> None:
    async with aiomysql.connect(port=port) as conn1:
        async with aiomysql.connect(port=port) as conn2:
            assert conn1.server_thread_id[0] + 1 == conn2.server_thread_id[0]


@pytest.mark.asyncio
async def test_replace_variables(
    session: MockSession, server: MysqlServer, connect: ConnectFixture
) -> None:
    session.echo = True

    with closing(await connect(user="levon_helm")) as conn:
        result = await query(conn, "SELECT CONNECTION_ID()")
        parts = result[0]["sql"].split(" ")
        assert parts[0] == "SELECT"
        int(parts[1])  # no error raised

        result = await query(conn, "SELECT @@version_comment")
        parts = result[0]["sql"].split(" ")
        assert parts[0] == "SELECT"
        assert parts[1] == "'mysql-mimic'"

        result = await query(conn, "SELECT CURRENT_USER")
        parts = result[0]["sql"].split(" ")
        assert parts[0] == "SELECT"
        assert parts[1] == "'levon_helm'"


@pytest.mark.asyncio
async def test_query_attributes(
    session: MockSession, server: MysqlServer, mysql_connector_conn: MySQLConnection
) -> None:
    session.echo = True

    for i, q in enumerate(
        [
            partial(query, conn=mysql_connector_conn),
            partial(query, conn=mysql_connector_conn, cursor_class=PreparedDictCursor),
        ]
    ):
        session.last_query_attrs = None
        sql = "SELECT 1"
        query_attrs = {
            "idx": i,
            "str": "foo",
            "int": 1,
            "float": 1.1,
        }
        result = await q(sql=sql, query_attributes=query_attrs)
        assert session.last_query_attrs == query_attrs
        assert result[0]["sql"] == sql



from contextlib import closing
from typing import Optional, Tuple, List, Dict

import pytest
from mysql.connector import DatabaseError
from mysql.connector.abstracts import MySQLConnectionAbstract
from mysql.connector.plugins.mysql_clear_password import MySQLClearPasswordAuthPlugin

from mysql_mimic import User, MysqlServer
from mysql_mimic.auth import (
    NativePasswordAuthPlugin,
    GullibleAuthPlugin,
    AbstractClearPasswordAuthPlugin,
    AuthPlugin,
    NoLoginAuthPlugin,
)
from tests.conftest import query, to_thread, MockSession, ConnectFixture

# mysql.connector throws an error if you try to use mysql_clear_password without SSL.
# That's silly, since SSL termination doesn't have to be handled by MySQL.
# But it's extra silly in tests.
MySQLClearPasswordAuthPlugin.requires_ssl = False
MySQLConnectionAbstract.is_secure = True  # pylint: disable=protected-access

SIMPLE_AUTH_USER = "levon_helm"

PASSWORD_AUTH_USER = "rick_danko"
PASSWORD_AUTH_PASSWORD = "nazareth"
PASSWORD_AUTH_OLD_PASSWORD = "cannonball"
PASSWORD_AUTH_PLUGIN = NativePasswordAuthPlugin.client_plugin_name


class TestPlugin(AbstractClearPasswordAuthPlugin):
    name = "test_plugin"

    async def check(self, username: str, password: str) -> Optional[str]:
        return username if username == password else None


TEST_PLUGIN_AUTH_USER = "garth_hudson"
TEST_PLUGIN_AUTH_PASSWORD = TEST_PLUGIN_AUTH_USER
TEST_PLUGIN_AUTH_PLUGIN = TestPlugin.client_plugin_name

NO_LOGIN_USER = "carmen_and_the_devil"
NO_LOGIN_PLUGIN = NoLoginAuthPlugin.name

UNKNOWN_PLUGIN_USER = "richard_manuel"
NO_PLUGIN_USER = "miss_moses"


@pytest.fixture(autouse=True)
def users() -> Dict[str, User]:
    return {
        SIMPLE_AUTH_USER: User(
            name=SIMPLE_AUTH_USER, auth_plugin=GullibleAuthPlugin.name
        ),
        PASSWORD_AUTH_USER: User(
            name=PASSWORD_AUTH_USER,
            auth_string=NativePasswordAuthPlugin.create_auth_string(
                PASSWORD_AUTH_PASSWORD
            ),
            old_auth_string=NativePasswordAuthPlugin.create_auth_string(
                PASSWORD_AUTH_OLD_PASSWORD
            ),
            auth_plugin=NativePasswordAuthPlugin.name,
        ),
        TEST_PLUGIN_AUTH_USER: User(
            name=TEST_PLUGIN_AUTH_USER,
            auth_string=TEST_PLUGIN_AUTH_PASSWORD,
            auth_plugin=TestPlugin.name,
        ),
        UNKNOWN_PLUGIN_USER: User(name=UNKNOWN_PLUGIN_USER, auth_plugin="unknown"),
        NO_PLUGIN_USER: User(name=NO_PLUGIN_USER, auth_plugin=NO_LOGIN_PLUGIN),
    }


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "auth_plugins,username,password,auth_plugin",
    [
        (
            [NativePasswordAuthPlugin()],
            PASSWORD_AUTH_USER,
            PASSWORD_AUTH_PASSWORD,
            PASSWORD_AUTH_PLUGIN,
        ),
        (
            [TestPlugin()],
            TEST_PLUGIN_AUTH_USER,
            TEST_PLUGIN_AUTH_PASSWORD,
            TEST_PLUGIN_AUTH_PLUGIN,
        ),
        ([GullibleAuthPlugin()], SIMPLE_AUTH_USER, None, None),
        ([TestPlugin(), GullibleAuthPlugin()], SIMPLE_AUTH_USER, None, None),
        (None, SIMPLE_AUTH_USER, None, None),
    ],
)
async def test_auth(
    server: MysqlServer,
    session: MockSession,
    connect: ConnectFixture,
    auth_plugins: List[AuthPlugin],
    username: str,
    password: Optional[str],
    auth_plugin: Optional[str],
) -> None:
    session.use_sqlite = True
    kwargs = {"user": username, "password": password, "auth_plugin": auth_plugin}
    kwargs = {k: v for k, v in kwargs.items() if v is not None}
    with closing(await connect(**kwargs)) as conn:
        assert await query(conn=conn, sql="SELECT USER() AS a") == [{"a": username}]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "auth_plugins",
    [
        [NativePasswordAuthPlugin()],
    ],
)
async def test_auth_secondary_password(
    server: MysqlServer,
    session: MockSession,
    connect: ConnectFixture,
    auth_plugins: List[AuthPlugin],
) -> None:
    session.use_sqlite = True
    with closing(
        await connect(
            user=PASSWORD_AUTH_USER,
            password=PASSWORD_AUTH_OLD_PASSWORD,
            auth_plugin=PASSWORD_AUTH_PLUGIN,
        )
    ) as conn:
        assert await query(conn=conn, sql="SELECT USER() AS a") == [
            {"a": PASSWORD_AUTH_USER}
        ]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "auth_plugins,user1,user2",
    [
        (
            [NativePasswordAuthPlugin(), TestPlugin()],
            (PASSWORD_AUTH_USER, PASSWORD_AUTH_PASSWORD, PASSWORD_AUTH_PLUGIN),
            (TEST_PLUGIN_AUTH_USER, TEST_PLUGIN_AUTH_PASSWORD, TEST_PLUGIN_AUTH_PLUGIN),
        ),
        (
            [GullibleAuthPlugin()],
            (TEST_PLUGIN_AUTH_USER, TEST_PLUGIN_AUTH_PASSWORD, TEST_PLUGIN_AUTH_PLUGIN),
            (PASSWORD_AUTH_USER, PASSWORD_AUTH_PASSWORD, PASSWORD_AUTH_PLUGIN),
        ),
        (
            [NativePasswordAuthPlugin()],
            (PASSWORD_AUTH_USER, PASSWORD_AUTH_PASSWORD, PASSWORD_AUTH_PLUGIN),
            (PASSWORD_AUTH_USER, PASSWORD_AUTH_PASSWORD, PASSWORD_AUTH_PLUGIN),
        ),
    ],
)
async def test_change_user(
    server: MysqlServer,
    session: MockSession,
    connect: ConnectFixture,
    auth_plugins: List[AuthPlugin],
    user1: Tuple[str, str, str],
    user2: Tuple[str, str, str],
) -> None:
    session.use_sqlite = True
    kwargs1 = {"user": user1[0], "password": user1[1], "auth_plugin": user1[2]}
    kwargs1 = {k: v for k, v in kwargs1.items() if v is not None}

    with closing(await connect(**kwargs1)) as conn:
        # mysql.connector doesn't have great support for COM_CHANGE_USER
        # Here, we have to manually override the auth plugin to use
        conn._auth_plugin = user2[2]  # pylint: disable=protected-access

        await to_thread(conn.cmd_change_user, username=user2[0], password=user2[1])
        assert await query(conn=conn, sql="SELECT USER() AS a") == [{"a": user2[0]}]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "auth_plugins,username,password,auth_plugin,msg",
    [
        ([GullibleAuthPlugin()], None, None, None, "User  does not exist"),
        (
            [TestPlugin()],
            PASSWORD_AUTH_USER,
            PASSWORD_AUTH_PASSWORD,
            PASSWORD_AUTH_PLUGIN,
            "Access denied",
        ),
        ([NoLoginAuthPlugin()], NO_PLUGIN_USER, None, None, "Access denied"),
        (
            [GullibleAuthPlugin(), NoLoginAuthPlugin()],
            NO_PLUGIN_USER,
            None,
            None,
            "Access denied",
        ),
    ],
)
async def test_access_denied(
    server: MysqlServer,
    session: MockSession,
    connect: ConnectFixture,
    auth_plugins: Optional[List[AuthPlugin]],
    username: Optional[str],
    password: Optional[str],
    auth_plugin: Optional[str],
    msg: str,
) -> None:
    session.use_sqlite = True
    kwargs = {"user": username, "password": password, "auth_plugin": auth_plugin}
    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    with pytest.raises(DatabaseError) as ctx:
        await connect(**kwargs)

    assert msg in str(ctx.value)



"""Implementation of the mysql server wire protocol"""
from mysql_mimic.auth import (
    User,
    IdentityProvider,
    NativePasswordAuthPlugin,
    NoLoginAuthPlugin,
    AuthPlugin,
)
from mysql_mimic.results import AllowedResult, ResultColumn, ResultSet
from mysql_mimic.session import Session
from mysql_mimic.server import MysqlServer



import pytest

from mysql_mimic.errors import MysqlError
from mysql_mimic.stream import MysqlStream


class MockReader:
    def __init__(self, data: bytes):
        self.data = data
        self.pos = 0

    async def read(self, n: int) -> bytes:
        start = self.pos
        end = self.pos + n
        self.pos = end

        return self.data[start:end]


class MockWriter:
    def __init__(self) -> None:
        self.data = b""

    def write(self, data: bytes) -> None:
        self.data += data

    async def drain(self) -> None:
        return


def test_seq() -> None:
    s1 = MysqlStream(reader=None, writer=None)  # type: ignore
    assert next(s1.seq) == 0
    assert next(s1.seq) == 1
    assert next(s1.seq) == 2
    s2 = MysqlStream(reader=None, writer=None)  # type: ignore
    assert next(s2.seq) == 0
    assert next(s2.seq) == 1
    assert next(s2.seq) == 2


def test_reset_seq() -> None:
    s = MysqlStream(reader=None, writer=None)  # type: ignore
    assert next(s.seq) == 0
    assert next(s.seq) == 1
    s.reset_seq()
    assert next(s.seq) == 0
    assert next(s.seq) == 1


@pytest.mark.asyncio
async def test_bad_seq_read() -> None:
    reader = MockReader(b"\x00\x00\x00\x01")
    s = MysqlStream(reader=reader, writer=None)  # type: ignore
    with pytest.raises(MysqlError):
        await s.read()


@pytest.mark.asyncio
async def test_empty_read() -> None:
    reader = MockReader(b"\x00\x00\x00\x00")
    s = MysqlStream(reader=reader, writer=None)  # type: ignore
    assert await s.read() == b""
    assert next(s.seq) == 1


@pytest.mark.asyncio
async def test_small_read() -> None:
    reader = MockReader(b"\x01\x00\x00\x00k")
    s = MysqlStream(reader=reader, writer=None)  # type: ignore
    assert await s.read() == b"k"
    assert next(s.seq) == 1


@pytest.mark.asyncio
async def test_medium_read() -> None:
    reader = MockReader(b"\xff\xff\x00\x00" + bytes(0xFFFF))
    s = MysqlStream(reader=reader, writer=None)  # type: ignore
    assert await s.read() == bytes(0xFFFF)
    assert next(s.seq) == 1


@pytest.mark.asyncio
async def test_edge_read() -> None:
    reader = MockReader(b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x00\x00\x00\x01")
    s = MysqlStream(reader=reader, writer=None)  # type: ignore
    assert await s.read() == bytes(0xFFFFFF)
    assert next(s.seq) == 2


@pytest.mark.asyncio
async def test_large_read() -> None:
    reader = MockReader(
        b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x06\x00\x00\x01" + b"kelsin"
    )
    s = MysqlStream(reader=reader, writer=None)  # type: ignore
    assert await s.read() == bytes(0xFFFFFF) + b"kelsin"
    assert next(s.seq) == 2


@pytest.mark.asyncio
async def test_empty_write() -> None:
    writer = MockWriter()
    s = MysqlStream(reader=None, writer=writer)  # type: ignore
    await s.write(b"")
    assert writer.data == b"\x00\x00\x00\x00"
    assert next(s.seq) == 1


@pytest.mark.asyncio
async def test_small_write() -> None:
    writer = MockWriter()
    s = MysqlStream(reader=None, writer=writer)  # type: ignore
    await s.write(b"kelsin")
    assert writer.data == b"\x06\x00\x00\x00kelsin"
    assert next(s.seq) == 1


@pytest.mark.asyncio
async def test_medium_write() -> None:
    writer = MockWriter()
    s = MysqlStream(reader=None, writer=writer)  # type: ignore
    await s.write(bytes(0xFFFF))
    assert writer.data == b"\xff\xff\x00\x00" + bytes(0xFFFF)
    assert next(s.seq) == 1


@pytest.mark.asyncio
async def test_edge_write() -> None:
    writer = MockWriter()
    s = MysqlStream(reader=None, writer=writer)  # type: ignore
    await s.write(bytes(0xFFFFFF))
    assert writer.data == b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x00\x00\x00\x01"
    assert next(s.seq) == 2


@pytest.mark.asyncio
async def test_large_write() -> None:
    writer = MockWriter()
    s = MysqlStream(reader=None, writer=writer)  # type: ignore
    await s.write(bytes(0xFFFFFF) + b"kelsin")
    assert (
        writer.data == b"\xff\xff\xff\x00" + bytes(0xFFFFFF) + b"\x06\x00\x00\x01kelsin"
    )
    assert next(s.seq) == 2



from typing import Dict, Any

import pytest

from mysql_mimic.charset import CharacterSet
from mysql_mimic.results import ResultSet
from mysql_mimic.admin import Admin
from mysql_mimic.variables import SystemVariables
from tests.conftest import MockSession


@pytest.fixture(scope="module")
def session() -> MockSession:
    return MockSession()


@pytest.fixture(scope="module")
def admin(session: MockSession) -> Admin:
    session.columns = {
        ("db", "table"): [
            {"name": "col1", "type": "INTEGER"},
            {"name": "col2", "type": "INTEGER"},
        ]
    }
    admin = Admin(connection_id=1, session=session, variables=SystemVariables())
    admin.database = "db"
    return admin


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cmd,result_length,col_length",
    [
        ("show columns from table", 2, 6),
        (
            " SHOW  EXTENDED  FULL  FIELDS  IN  `table`  IN  `db` ",
            2,
            9,
        ),
        ("show columns from table like '%2'", 1, 6),
        ("show columns from table like 'col1'", 1, 6),
        ("show columns from table like 'col%'", 2, 6),
        ("show columns from table like '%col%'", 2, 6),
        ("show columns from table like '_ol2'", 1, 6),
        ("show tables from db", 1, 1),
        ("show full tables from db", 1, 2),
        ("show tables from db like 'table'", 1, 1),
        ("show tables from db like 'x'", 0, 1),
        ("show databases", 1, 1),
        ("show schemas", 1, 1),
        ("show databases like 'db'", 1, 1),
        ("show databases like 'x'", 0, 1),
    ],
)
async def test_parse_show(
    session: MockSession,
    admin: Admin,
    cmd: str,
    result_length: int,
    col_length: int,
) -> None:
    result = await admin.parse(cmd)
    assert isinstance(result, ResultSet)
    assert len(list(result.rows)) == result_length
    assert len(result.columns) == col_length


@pytest.mark.asyncio
async def test_parse_show_index(admin: Admin) -> None:
    cmd = "show index from table"
    result_length = 0
    col_length = 15
    result = await admin.parse(cmd)
    assert isinstance(result, ResultSet)
    assert len(list(result.rows)) == result_length
    assert len(result.columns) == col_length


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cmd,expected",
    [
        (
            "show variables",
            {
                "version_comment": "mysql-mimic",
                "version": "8.0.29",
                "character_set_client": "utf8mb4",
                "character_set_results": "utf8mb4",
                "character_set_server": "utf8mb4",
                "collation_server": "utf8mb4_general_ci",
                "collation_database": "utf8mb4_general_ci",
                "lower_case_table_names": 0,
                "sql_mode": "",
                "transaction_isolation": "READ-COMMITTED",
                "external_user": "",
            },
        ),
        (
            "SHOW  SESSION  VARIABLES  LIKE 'version_comment'",
            {"version_comment": "mysql-mimic"},
        ),
        (
            "SHOW  SESSION  VARIABLES  LIKE 'version_%'",
            {"version_comment": "mysql-mimic"},
        ),
        ("SET @@version_comment = 'hello'", {}),
        (
            "SHOW  SESSION  VARIABLES  LIKE 'version_comment'",
            {"version_comment": "hello"},
        ),
    ],
)
async def test_parse_show_variables(
    admin: Admin, cmd: str, expected: Dict[str, Any]
) -> None:
    result = await admin.parse(cmd)
    assert isinstance(result, ResultSet)
    rows = {r[0]: r[1] for r in result.rows}
    assert rows == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cmd,expected_charset,expected_collation",
    [
        (
            "SET NAMES utf8",
            "utf8",
            "utf8_general_ci",
        ),
        (
            "SET NAMES big5 DEFAULT",
            "big5",
            "big5_chinese_ci",
        ),
        (
            "SET NAMES utf8 COLLATE utf8mb4_bin",
            "utf8",
            "utf8mb4_bin",
        ),
        (
            "set  names  'big5'  collate  'big5_chinese_ci'",
            "big5",
            "big5_chinese_ci",
        ),
    ],
)
async def test_parse_set_names(
    admin: Admin, cmd: str, expected_charset: str, expected_collation: str
) -> None:
    await admin.parse(cmd)
    assert admin.vars["character_set_client"] == expected_charset
    assert admin.vars["character_set_connection"] == expected_charset
    assert admin.vars["character_set_results"] == expected_charset
    assert admin.vars["collation_connection"] == expected_collation
    assert admin.vars.client_charset == CharacterSet[expected_charset]
    assert admin.vars.server_charset == CharacterSet[expected_charset]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "cmd,expected",
    [
        ("SET version_comment = 'b'", "b"),
        ("SET @@version_comment = ON", True),
        ("SET  @@SESSION.version_comment  = OFF", False),
        ("set version_comment = DEFAULT", "mysql-mimic"),
        ("set version_comment = 1", 1),
        ("set version_comment = 2.3", 2.3),
        ("set version_comment = NULL", "mysql-mimic"),
    ],
)
async def test_parse_set_variables(admin: Admin, cmd: str, expected: str) -> None:
    await admin.parse(cmd)
    assert admin.vars["version_comment"] == expected



import asyncio
import struct
from ssl import SSLContext

from mysql_mimic.errors import MysqlError, ErrorCode
from mysql_mimic.types import uint_3, uint_1
from mysql_mimic.utils import seq


class ConnectionClosed(Exception):
    pass


class MysqlStream:
    def __init__(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        self.reader = reader
        self.writer = writer
        self.seq = seq(256)

    async def read(self) -> bytes:
        data = b""
        while True:
            header = await self.reader.read(4)

            if not header:
                raise ConnectionClosed()

            i = struct.unpack("<I", header)[0]
            payload_length = i & 0x00FFFFFF
            sequence_id = (i & 0xFF000000) >> 24

            expected = next(self.seq)
            if sequence_id != expected:
                raise MysqlError(
                    f"Expected seq({expected}) got seq({sequence_id})",
                    ErrorCode.MALFORMED_PACKET,
                )

            if payload_length == 0:
                return data

            data += await self.reader.read(payload_length)

            if payload_length < 0xFFFFFF:
                return data

    async def write(self, data: bytes) -> None:
        while True:
            # Grab first 0xFFFFFF bytes to send
            payload = data[:0xFFFFFF]
            data = data[0xFFFFFF:]

            payload_length = uint_3(len(payload))
            sequence_id = uint_1(next(self.seq))

            self.writer.write(payload_length + sequence_id + payload)
            await self.writer.drain()

            # We are done unless len(send) == 0xFFFFFF
            if len(payload) != 0xFFFFFF:
                return

    def reset_seq(self) -> None:
        self.seq.reset()

    async def start_tls(self, ssl: SSLContext) -> None:
        transport = self.writer.transport
        protocol = transport.get_protocol()
        loop = asyncio.get_event_loop()
        new_transport = await loop.start_tls(
            transport=transport,
            protocol=protocol,
            sslcontext=ssl,
            server_side=True,
        )

        # This seems to be the easiest way to wrap the socket created by asyncio
        self.writer._transport = new_transport  # type: ignore # pylint: disable=protected-access
        self.reader._transport = new_transport  # type: ignore # pylint: disable=protected-access



import io
from dataclasses import dataclass, field
from typing import Optional, Dict, Any, Sequence, Callable, Tuple, List, Union

from mysql_mimic.charset import Collation, CharacterSet
from mysql_mimic.errors import ErrorCode, get_sqlstate, MysqlError
from mysql_mimic.prepared import PreparedStatement, REGEX_PARAM
from mysql_mimic.results import NullBitmap, ResultColumn
from mysql_mimic.types import (
    Capabilities,
    uint_2,
    uint_1,
    str_null,
    uint_4,
    str_fixed,
    read_uint_4,
    read_str_null,
    read_str_fixed,
    read_uint_1,
    read_str_len,
    read_uint_len,
    str_len,
    str_rest,
    uint_len,
    ColumnType,
    read_int_1,
    read_int_2,
    read_uint_2,
    read_int_4,
    read_uint_8,
    read_int_8,
    read_float,
    read_double,
    ResultsetMetadata,
    ColumnDefinition,
    ComStmtExecuteFlags,
    peek,
    ServerStatus,
)


@dataclass
class SSLRequest:
    max_packet_size: int
    capabilities: Capabilities
    client_charset: CharacterSet


@dataclass
class HandshakeResponse41:
    max_packet_size: int
    capabilities: Capabilities
    client_charset: CharacterSet
    username: str
    auth_response: bytes
    connect_attrs: Dict[str, str] = field(default_factory=dict)
    database: Optional[str] = None
    client_plugin: Optional[str] = None
    zstd_compression_level: int = 0


@dataclass
class ComChangeUser:
    username: str
    auth_response: bytes
    database: str
    client_charset: Optional[CharacterSet] = None
    client_plugin: Optional[str] = None
    connect_attrs: Dict[str, str] = field(default_factory=dict)


@dataclass
class ComQuery:
    sql: str
    query_attrs: Dict[str, str]


@dataclass
class ComStmtSendLongData:
    stmt_id: int
    param_id: int
    data: bytes


@dataclass
class ComStmtExecute:
    sql: str
    query_attrs: Dict[str, str]
    stmt: PreparedStatement
    use_cursor: bool


@dataclass
class ComStmtFetch:
    stmt_id: int
    num_rows: int


@dataclass
class ComStmtReset:
    stmt_id: int


@dataclass
class ComStmtClose:
    stmt_id: int


def make_ok(
    capabilities: Capabilities,
    status_flags: ServerStatus,
    eof: bool = False,
    affected_rows: int = 0,
    last_insert_id: int = 0,
    warnings: int = 0,
    flags: int = 0,
) -> bytes:
    parts = [
        uint_1(0) if not eof else uint_1(0xFE),
        uint_len(affected_rows),
        uint_len(last_insert_id),
    ]

    if Capabilities.CLIENT_PROTOCOL_41 in capabilities:
        parts.append(uint_2(status_flags | flags))
        parts.append(uint_2(warnings))
    elif Capabilities.CLIENT_TRANSACTIONS in capabilities:
        parts.append(uint_2(status_flags | flags))

    return _concat(*parts)


def make_eof(
    capabilities: Capabilities,
    status_flags: ServerStatus,
    warnings: int = 0,
    flags: int = 0,
) -> bytes:
    parts = [uint_1(0xFE)]

    if Capabilities.CLIENT_PROTOCOL_41 in capabilities:
        parts.append(uint_2(warnings))
        parts.append(uint_2(status_flags | flags))

    return _concat(*parts)


def make_error(
    capabilities: Capabilities,
    server_charset: CharacterSet,
    msg: Any,
    code: ErrorCode = ErrorCode.UNKNOWN_ERROR,
) -> bytes:
    parts = [uint_1(0xFF), uint_2(code)]

    if Capabilities.CLIENT_PROTOCOL_41 in capabilities:
        parts.append(str_fixed(1, b"#"))
        parts.append(str_fixed(5, get_sqlstate(code)))

    parts.append(str_rest(server_charset.encode(str(msg))))

    return _concat(*parts)


def make_handshake_v10(
    capabilities: Capabilities,
    server_charset: CharacterSet,
    server_version: str,
    connection_id: int,
    auth_data: bytes,
    status_flags: ServerStatus,
    auth_plugin_name: str,
) -> bytes:
    auth_plugin_data_len = (
        len(auth_data) if Capabilities.CLIENT_PLUGIN_AUTH in capabilities else 0
    )

    parts = [
        uint_1(10),  # protocol version
        str_null(server_charset.encode(server_version)),  # server version
        uint_4(connection_id),  # connection ID
        str_null(auth_data[:8]),  # plugin data
        uint_2(capabilities & 0xFFFF),  # lower capabilities flag
        uint_1(server_charset),  # lower character set
        uint_2(status_flags),  # server status flag
        uint_2(capabilities >> 16),  # higher capabilities flag
        uint_1(auth_plugin_data_len),
        str_fixed(10, bytes(10)),  # reserved
        str_fixed(max(13, auth_plugin_data_len - 8), auth_data[8:]),
    ]
    if Capabilities.CLIENT_PLUGIN_AUTH in capabilities:
        parts.append(str_null(server_charset.encode(auth_plugin_name)))
    return _concat(*parts)


def parse_handshake_response(
    capabilities: Capabilities, data: bytes
) -> Union[HandshakeResponse41, SSLRequest]:
    r = io.BytesIO(data)

    client_capabilities = Capabilities(read_uint_4(r))

    capabilities = capabilities & client_capabilities

    max_packet_size = read_uint_4(r)
    client_charset = Collation(read_uint_1(r)).charset
    read_str_fixed(r, 23)

    if not peek(r):
        return SSLRequest(
            max_packet_size=max_packet_size,
            capabilities=capabilities,
            client_charset=client_charset,
        )

    username = client_charset.decode(read_str_null(r))

    if Capabilities.CLIENT_PLUGIN_AUTH_LENENC_CLIENT_DATA in capabilities:
        auth_response = read_str_len(r)
    else:
        l_auth_response = read_uint_1(r)
        auth_response = read_str_fixed(r, l_auth_response)

    response = HandshakeResponse41(
        max_packet_size=max_packet_size,
        capabilities=capabilities,
        client_charset=client_charset,
        username=username,
        auth_response=auth_response,
    )

    if Capabilities.CLIENT_CONNECT_WITH_DB in capabilities:
        response.database = client_charset.decode(read_str_null(r))

    if Capabilities.CLIENT_PLUGIN_AUTH in capabilities:
        response.client_plugin = client_charset.decode(read_str_null(r))

    if Capabilities.CLIENT_CONNECT_ATTRS in capabilities:
        response.connect_attrs = _read_connect_attrs(r, client_charset)

    if Capabilities.CLIENT_ZSTD_COMPRESSION_ALGORITHM in capabilities:
        response.zstd_compression_level = read_uint_1(r)

    return response


def parse_handshake_response_41(
    capabilities: Capabilities, data: bytes
) -> HandshakeResponse41:
    response = parse_handshake_response(capabilities, data)
    assert isinstance(response, HandshakeResponse41)
    return response


def make_auth_more_data(data: bytes) -> bytes:
    return _concat(uint_1(1), str_rest(data))  # status tag


def make_auth_switch_request(
    server_charset: CharacterSet, plugin_name: str, plugin_provided_data: bytes
) -> bytes:
    return _concat(
        uint_1(254),  # status tag
        str_null(server_charset.encode(plugin_name)),
        str_rest(plugin_provided_data),
    )


def parse_com_change_user(
    capabilities: Capabilities, client_charset: CharacterSet, data: bytes
) -> ComChangeUser:
    r = io.BytesIO(data)
    username = client_charset.decode(read_str_null(r))
    if Capabilities.CLIENT_SECURE_CONNECTION in capabilities:
        l_auth_response = read_uint_1(r)
        auth_response = read_str_fixed(r, l_auth_response)
    else:
        auth_response = read_str_null(r)
    database = client_charset.decode(read_str_null(r))

    response = ComChangeUser(
        username=username, auth_response=auth_response, database=database
    )

    if peek(r):  # more data available
        if Capabilities.CLIENT_PROTOCOL_41 in capabilities:
            client_charset = Collation(read_uint_2(r)).charset
            response.client_charset = client_charset
        if Capabilities.CLIENT_PLUGIN_AUTH in capabilities:
            response.client_plugin = client_charset.decode(read_str_null(r))
        if Capabilities.CLIENT_CONNECT_ATTRS in capabilities:
            response.connect_attrs = _read_connect_attrs(r, client_charset)

    return response


def parse_com_query(
    capabilities: Capabilities, client_charset: CharacterSet, data: bytes
) -> ComQuery:
    r = io.BytesIO(data)

    if Capabilities.CLIENT_QUERY_ATTRIBUTES in capabilities:
        parameter_count = read_uint_len(r)
        read_uint_len(r)  # parameter_set_count. Always 1.
        query_attrs = {
            k: v
            for k, v in _read_params(capabilities, client_charset, r, parameter_count)
            if k is not None
        }
    else:
        query_attrs = {}

    sql = r.read().decode(client_charset.codec)

    return ComQuery(
        sql=sql,
        query_attrs=query_attrs,
    )


def make_column_count(capabilities: Capabilities, column_count: int) -> bytes:
    parts = []

    if Capabilities.CLIENT_OPTIONAL_RESULTSET_METADATA in capabilities:
        parts.append(uint_1(ResultsetMetadata.RESULTSET_METADATA_FULL))

    parts.append(uint_len(column_count))

    return _concat(*parts)


# pylint: disable=too-many-arguments
def make_column_definition_41(
    server_charset: CharacterSet,
    schema: Optional[str] = None,
    table: Optional[str] = None,
    org_table: Optional[str] = None,
    name: Optional[str] = None,
    org_name: Optional[str] = None,
    character_set: CharacterSet = CharacterSet.utf8mb4,
    column_length: int = 256,
    column_type: ColumnType = ColumnType.VARCHAR,
    flags: ColumnDefinition = ColumnDefinition(0),
    decimals: int = 0,
) -> bytes:
    schema = schema or ""
    table = table or ""
    org_table = org_table or table
    name = name or ""
    org_name = org_name or name

    return _concat(
        str_len(b"def"),
        str_len(server_charset.encode(schema)),
        str_len(server_charset.encode(table)),
        str_len(server_charset.encode(org_table)),
        str_len(server_charset.encode(name)),
        str_len(server_charset.encode(org_name)),
        uint_len(0x0C),  # Length of the following fields
        uint_2(character_set),
        uint_4(column_length),
        uint_1(column_type),
        uint_2(flags),
        uint_1(decimals),
        uint_2(0),  # filler
    )


def make_text_resultset_row(
    row: Sequence[Any], columns: Sequence[ResultColumn]
) -> bytes:
    parts = []

    for value, column in zip(row, columns):
        if value is None:
            parts.append(b"\xfb")
        else:
            text = column.text_encode(value)
            parts.append(str_len(text))

    return _concat(*parts)


def make_com_stmt_prepare_ok(statement: PreparedStatement) -> bytes:
    return _concat(
        uint_1(0),  # OK
        uint_4(statement.stmt_id),
        uint_2(0),  # number of columns
        uint_2(statement.num_params),
        uint_1(0),  # filler
        uint_2(0),  # number of warnings
    )


def parse_com_stmt_send_long_data(data: bytes) -> ComStmtSendLongData:
    r = io.BytesIO(data)
    return ComStmtSendLongData(
        stmt_id=read_uint_4(r),
        param_id=read_uint_2(r),
        data=r.read(),
    )


def parse_com_stmt_execute(
    capabilities: Capabilities,
    client_charset: CharacterSet,
    data: bytes,
    get_stmt: Callable[[int], PreparedStatement],
) -> ComStmtExecute:
    r = io.BytesIO(data)
    stmt_id = read_uint_4(r)
    stmt = get_stmt(stmt_id)
    use_cursor, param_count_available = _read_cursor_flags(r)
    read_uint_4(r)  # iteration count. Always 1.
    sql, query_attrs = _interpolate_params(
        capabilities, client_charset, r, stmt, param_count_available
    )
    return ComStmtExecute(
        sql=sql,
        query_attrs=query_attrs,
        stmt=stmt,
        use_cursor=use_cursor,
    )


def make_binary_resultrow(row: Sequence[Any], columns: Sequence[ResultColumn]) -> bytes:
    column_count = len(row)

    null_bitmap = NullBitmap.new(column_count, offset=2)

    values = []
    for i, (val, col) in enumerate(zip(row, columns)):
        if val is None:
            null_bitmap.flip(i)
        else:
            values.append(col.binary_encode(val))

    values_data = b"".join(values)

    null_bitmap_data = bytes(null_bitmap)

    return b"".join(
        [
            uint_1(0),  # packet header
            str_fixed(len(null_bitmap_data), null_bitmap_data),
            str_fixed(len(values_data), values_data),
        ]
    )


def parse_handle_stmt_fetch(data: bytes) -> ComStmtFetch:
    r = io.BytesIO(data)
    return ComStmtFetch(
        stmt_id=read_uint_4(r),
        num_rows=read_uint_4(r),
    )


def parse_com_stmt_reset(data: bytes) -> ComStmtReset:
    r = io.BytesIO(data)
    return ComStmtReset(stmt_id=read_uint_4(r))


def parse_com_stmt_close(data: bytes) -> ComStmtClose:
    r = io.BytesIO(data)
    return ComStmtClose(stmt_id=read_uint_4(r))


def _read_cursor_flags(reader: io.BytesIO) -> Tuple[bool, bool]:
    flags = ComStmtExecuteFlags(read_uint_1(reader))
    param_count_available = ComStmtExecuteFlags.PARAMETER_COUNT_AVAILABLE in flags

    if ComStmtExecuteFlags.CURSOR_TYPE_READ_ONLY in flags:
        return True, param_count_available
    if ComStmtExecuteFlags.CURSOR_TYPE_NO_CURSOR in flags:
        return False, param_count_available
    raise MysqlError(f"Unsupported cursor flags: {flags}", ErrorCode.NOT_SUPPORTED_YET)


def _interpolate_params(
    capabilities: Capabilities,
    client_charset: CharacterSet,
    reader: io.BytesIO,
    stmt: PreparedStatement,
    param_count_available: bool,
) -> Tuple[str, Dict[str, str]]:
    sql = stmt.sql
    query_attrs = {}
    parameter_count = stmt.num_params

    if stmt.num_params > 0 or (
        Capabilities.CLIENT_QUERY_ATTRIBUTES in capabilities and param_count_available
    ):
        if Capabilities.CLIENT_QUERY_ATTRIBUTES in capabilities:
            parameter_count = read_uint_len(reader)

    if parameter_count > 0:
        # When there are query attributes, they are combined with statement parameters.
        # The statement parameters will be first, query attributes second.
        params = _read_params(
            capabilities, client_charset, reader, parameter_count, stmt.param_buffers
        )

        for _, value in params[: stmt.num_params]:
            sql = REGEX_PARAM.sub(_encode_param_as_sql(value), sql, 1)

        query_attrs = {k: v for k, v in params[stmt.num_params :] if k is not None}

    return sql, query_attrs


def _encode_param_as_sql(param: Any) -> str:
    if isinstance(param, str):
        return f"'{param}'"
    if param is None:
        return "NULL"
    if param is True:
        return "TRUE"
    if param is False:
        return "FALSE"
    return str(param)


def _read_params(
    capabilities: Capabilities,
    client_charset: CharacterSet,
    reader: io.BytesIO,
    parameter_count: int,
    buffers: Optional[Dict[int, bytearray]] = None,
) -> Sequence[Tuple[Optional[str], Any]]:
    """
    Read parameters from a stream.

    This is intended for reading query attributes from COM_QUERY and parameters from COM_STMT_EXECUTE.

    Returns:
        Name/value pairs. Name is None if there is no name, which is the case for stmt_execute, which
        combines statement parameters and query attributes.
    """
    params: List[Tuple[Optional[str], Any]] = []

    if parameter_count:
        null_bitmap = NullBitmap.from_buffer(reader, parameter_count)
        new_params_bound_flag = read_uint_1(reader)

        if not new_params_bound_flag:
            raise MysqlError(
                "Server requires the new-params-bound-flag to be set",
                ErrorCode.NOT_SUPPORTED_YET,
            )

        param_types = []

        for i in range(parameter_count):
            param_type, unsigned = _read_param_type(reader)

            if Capabilities.CLIENT_QUERY_ATTRIBUTES in capabilities:
                # Only query attributes have names
                # Statement parameters will have an empty name, e.g. b"\x00"
                param_name = client_charset.decode(read_str_len(reader))
            else:
                param_name = ""

            param_types.append((param_name, param_type, unsigned))

        for i, (param_name, param_type, unsigned) in enumerate(param_types):
            if null_bitmap.is_flipped(i):
                params.append((param_name, None))
            elif buffers and i in buffers:
                params.append((param_name, client_charset.decode(buffers[i])))
            else:
                params.append(
                    (
                        param_name,
                        _read_param_value(client_charset, reader, param_type, unsigned),
                    )
                )

    return params


def _read_param_type(reader: io.BytesIO) -> Tuple[ColumnType, bool]:
    param_type = ColumnType(read_uint_1(reader))
    is_unsigned = (read_uint_1(reader) & 0x80) > 0
    return param_type, is_unsigned


def _read_param_value(
    client_charset: CharacterSet,
    reader: io.BytesIO,
    param_type: ColumnType,
    unsigned: bool,
) -> Any:
    if param_type in {
        ColumnType.VARCHAR,
        ColumnType.VAR_STRING,
        ColumnType.STRING,
        ColumnType.BLOB,
        ColumnType.TINY_BLOB,
        ColumnType.MEDIUM_BLOB,
        ColumnType.LONG_BLOB,
    }:
        val = read_str_len(reader)
        return client_charset.decode(val)

    if param_type == ColumnType.TINY:
        return (read_uint_1 if unsigned else read_int_1)(reader)

    if param_type == ColumnType.BOOL:
        return read_uint_1(reader)

    if param_type in {ColumnType.SHORT, ColumnType.YEAR}:
        return (read_uint_2 if unsigned else read_int_2)(reader)

    if param_type in {ColumnType.LONG, ColumnType.INT24}:
        return (read_uint_4 if unsigned else read_int_4)(reader)

    if param_type == ColumnType.LONGLONG:
        return (read_uint_8 if unsigned else read_int_8)(reader)

    if param_type == ColumnType.FLOAT:
        return read_float(reader)

    if param_type == ColumnType.DOUBLE:
        return read_double(reader)

    if param_type == ColumnType.NULL:
        return None

    raise MysqlError(
        f"Unsupported parameter type: {param_type}", ErrorCode.NOT_SUPPORTED_YET
    )


def _read_connect_attrs(
    reader: io.BytesIO, client_charset: CharacterSet
) -> Dict[str, str]:
    connect_attrs = {}
    total_l = read_uint_len(reader)

    while total_l > 0:
        key = read_str_len(reader)
        value = read_str_len(reader)
        connect_attrs[client_charset.decode(key)] = client_charset.decode(value)

        item_l = len(str_len(key) + str_len(value))
        total_l -= item_l
    return connect_attrs


def _concat(*parts: bytes) -> bytes:
    return b"".join(parts)



from typing import Any

import pytest

from mysql_mimic.errors import MysqlError
from mysql_mimic.results import ensure_result_set


@pytest.mark.parametrize(
    "result",
    [
        [1, 2],
        ([[1, 2]], ["a", "b"], ["a", "b"]),
        ([[1, 2]], [1, 2]),
    ],
)
def test_ensure_result_set__invalid(result: Any) -> None:
    with pytest.raises(MysqlError):
        ensure_result_set(result)



import os
from contextlib import closing
from ssl import SSLContext, PROTOCOL_TLS_SERVER

import pytest

from tests.conftest import MockSession, ConnectFixture, query

CURRENT_DIR = os.path.dirname(__file__)
SSL_CERT = os.path.join(CURRENT_DIR, "fixtures/certificate.pem")
SSL_KEY = os.path.join(CURRENT_DIR, "fixtures/key.pem")


@pytest.fixture(autouse=True)
def ssl() -> SSLContext:
    sslcontext = SSLContext(PROTOCOL_TLS_SERVER)
    sslcontext.load_cert_chain(certfile=SSL_CERT, keyfile=SSL_KEY)
    return sslcontext


@pytest.mark.asyncio
@pytest.mark.skipif(
    os.getenv("CI") == "true",
    reason="""
There seems to be a sequencing issue with the asyncio server starting tls.
https://stackoverflow.com/questions/74187508/sequencing-issue-when-upgrading-python-asyncio-connection-to-tls
Didn't have time to debug this, so skipping in CI for now.
""",
)
async def test_ssl(
    session: MockSession,
    connect: ConnectFixture,
    port: int,
) -> None:
    session.use_sqlite = True

    with closing(await connect()) as conn:
        results = await query(conn, "SELECT 1 AS a")
        assert results == [{"a": 1}]



from collections import UserDict
from typing import Dict, Any

from mysql_mimic.charset import CharacterSet, Collation

DEFAULT_VARIABLES = {
    "version": "8.0.29",
    "version_comment": "mysql-mimic",
    "character_set_client": CharacterSet.utf8mb4.name,
    "character_set_results": CharacterSet.utf8mb4.name,
    "character_set_server": CharacterSet.utf8mb4.name,
    "collation_server": Collation.utf8mb4_general_ci.name,
    "collation_database": Collation.utf8mb4_general_ci.name,
    "transaction_isolation": "READ-COMMITTED",
    "sql_mode": "",
    "lower_case_table_names": 0,
    "external_user": "",
}


class SystemVariables(UserDict):
    def __init__(self, defaults: Dict[str, Any] = None):
        self.defaults = defaults or DEFAULT_VARIABLES
        super().__init__(self.defaults)

    @property
    def server_charset(self) -> CharacterSet:
        return CharacterSet[self.get("character_set_results", "utf8mb4")]

    @server_charset.setter
    def server_charset(self, val: CharacterSet) -> None:
        self["character_set_results"] = val.name

    @property
    def client_charset(self) -> CharacterSet:
        return CharacterSet[self.get("character_set_client", "utf8mb4")]

    @client_charset.setter
    def client_charset(self, val: CharacterSet) -> None:
        self["character_set_client"] = val.name

    @property
    def external_user(self) -> str:
        return self["external_user"]

    @external_user.setter
    def external_user(self, val: str) -> None:
        self["external_user"] = val

    @property
    def mysql_version(self) -> str:
        return self["version"]



from setuptools import setup, find_packages

# Import __version__
exec(open("mysql_mimic/version.py").read())

setup(
    name="mysql-mimic",
    version=__version__,
    description="A python implementation of the mysql server protocol",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/kelsin/mysql-mimic",
    author="Christopher Giroir",
    author_email="kelsin@valefor.com",
    license="MIT",
    packages=find_packages(),
    python_requires=">=3.6",
    extras_require={
        "dev": [
            "aiomysql",
            "mypy",
            "mysql-connector-python",
            "black",
            "coverage",
            "pylint",
            "pytest",
            "pytest-asyncio",
            "sphinx",
            "sqlalchemy",
            "twine",
            "wheel",
        ],
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Programming Language :: SQL",
        "Programming Language :: Python :: 3 :: Only",
    ],
)



from __future__ import annotations

import io
from copy import copy
from hashlib import sha1
import logging
import secrets
from dataclasses import dataclass
from typing import Optional, Dict, AsyncGenerator, Union, Tuple, Sequence

from mysql_mimic.types import read_str_null
from mysql_mimic.utils import xor

logger = logging.getLogger(__name__)


# Many authentication plugins don't need to send any sort of challenge/nonce.
FILLER = b"0" * 20 + b"\x00"


@dataclass
class Forbidden:
    msg: Optional[str] = None


@dataclass
class Success:
    authenticated_as: str


@dataclass
class User:
    name: str
    auth_string: Optional[str] = None
    auth_plugin: Optional[str] = None

    # For plugins that support a primary and secondary password
    # This is helpful for zero-downtime password rotation
    old_auth_string: Optional[str] = None


@dataclass
class AuthInfo:
    username: str
    data: bytes
    user: User
    connect_attrs: Dict[str, str]
    client_plugin_name: Optional[str]
    handshake_auth_data: Optional[bytes]
    handshake_plugin_name: str

    def copy(self, data: bytes) -> AuthInfo:
        new = copy(self)
        new.data = data
        return new


Decision = Union[Success, Forbidden, bytes]
AuthState = AsyncGenerator[Decision, AuthInfo]


class AuthPlugin:
    """
    Abstract base class for authentication plugins.
    """

    name = ""
    client_plugin_name: Optional[str] = None  # None means any

    async def auth(self, auth_info: Optional[AuthInfo] = None) -> AuthState:
        """
        Create an async generator that drives the authentication lifecycle.

        This should either yield `bytes`, in which case an AuthMoreData packet is sent to the client,
        or a `Success` or `Forbidden` instance, in which case authentication is complete.

        Args:
             auth_info: This is None if authentication is starting from the optimistic handshake.
        """
        yield Forbidden()

    async def start(
        self, auth_info: Optional[AuthInfo] = None
    ) -> Tuple[Decision, AuthState]:
        state = self.auth(auth_info)
        data = await state.__anext__()
        return data, state


class GullibleAuthPlugin(AuthPlugin):
    """
    Custom plugin that naively accepts whatever username a client provides.
    """

    name = "mysql_mimic_gullible"

    async def auth(self, auth_info: Optional[AuthInfo] = None) -> AuthState:
        if not auth_info:
            auth_info = yield FILLER
        yield Success(authenticated_as=auth_info.username)


class AbstractClearPasswordAuthPlugin(AuthPlugin):
    """
    Abstract class for implementing the server-side of the standard client plugin "mysql_clear_password".
    """

    name = "abstract_mysql_clear_password"
    client_plugin_name = "mysql_clear_password"

    async def auth(self, auth_info: Optional[AuthInfo] = None) -> AuthState:
        if not auth_info:
            auth_info = yield FILLER

        r = io.BytesIO(auth_info.data)
        password = read_str_null(r).decode()
        authenticated_as = await self.check(auth_info.username, password)
        if authenticated_as is not None:
            yield Success(authenticated_as)
        else:
            yield Forbidden()

    async def check(self, username: str, password: str) -> Optional[str]:
        return username


class NativePasswordAuthPlugin(AuthPlugin):
    """
    Standard plugin that uses a password hashing method.

    The client hashed the password using a nonce provided by the server, so the
    password can't be snooped on the network.

    Furthermore, thanks to some clever hashing techniques, knowing the hash stored in the
    user database isn't enough to authenticate as that user.
    """

    name = "mysql_native_password"
    client_plugin_name = "mysql_native_password"

    async def auth(self, auth_info: Optional[AuthInfo] = None) -> AuthState:
        if (
            auth_info
            and auth_info.handshake_plugin_name == self.name
            and auth_info.handshake_auth_data
        ):
            # mysql_native_password can reuse the nonce from the initial handshake
            nonce = auth_info.handshake_auth_data.rstrip(b"\x00")
        else:
            nonce = secrets.token_bytes(20)
            # The mysql command line client expects a null terminating byte
            auth_info = yield nonce + b"\x00"

        user = auth_info.user
        if self.password_matches(user=user, scramble=auth_info.data, nonce=nonce):
            yield Success(user.name)
        else:
            yield Forbidden()

    def empty_password_quickpath(self, user: User, scramble: bytes) -> bool:
        return not scramble and not user.auth_string

    def password_matches(self, user: User, scramble: bytes, nonce: bytes) -> bool:
        return (
            self.empty_password_quickpath(user, scramble)
            or self.verify_scramble(user.auth_string, scramble, nonce)
            or self.verify_scramble(user.old_auth_string, scramble, nonce)
        )

    def verify_scramble(
        self, auth_string: Optional[str], scramble: bytes, nonce: bytes
    ) -> bool:
        # From docs,
        # response.data should be:
        #   SHA1(password) XOR SHA1("20-bytes random data from server" <concat> SHA1(SHA1(password)))
        # auth_string should be:
        #   SHA1(SHA1(password))
        sha1_sha1_password = bytes.fromhex(auth_string or "")
        sha1_sha1_with_nonce = sha1(nonce + sha1_sha1_password).digest()
        rcvd_sha1_password = xor(scramble, sha1_sha1_with_nonce)
        return sha1(rcvd_sha1_password).digest() == sha1_sha1_password

    @classmethod
    def create_auth_string(cls, password: str) -> str:
        return sha1(sha1(password.encode("utf-8")).digest()).hexdigest()


class NoLoginAuthPlugin(AuthPlugin):
    """
    Standard plugin that prevents all clients from direct login.

    This is useful for user accounts that can only be accessed by proxy authentication.
    """

    name = "mysql_no_login"

    async def auth(self, auth_info: Optional[AuthInfo] = None) -> AuthState:
        if not auth_info:
            _ = yield FILLER
        yield Forbidden()


class IdentityProvider:
    """
    Abstract base class for an identity provider.

    An identity provider tells the server with authentication plugins to make
    available to clients and how to retrieve users.
    """

    def get_plugins(self) -> Sequence[AuthPlugin]:
        return [NativePasswordAuthPlugin(), NoLoginAuthPlugin()]

    async def get_user(self, username: str) -> Optional[User]:
        return None

    def get_default_plugin(self) -> AuthPlugin:
        return self.get_plugins()[0]

    def get_plugin(self, name: str) -> Optional[AuthPlugin]:
        try:
            return next(p for p in self.get_plugins() if p.name == name)
        except StopIteration:
            return None


class SimpleIdentityProvider(IdentityProvider):
    """
    Simple identity provider implementation that naively accepts whatever username a client provides.
    """

    def get_plugins(self) -> Sequence[AuthPlugin]:
        return [GullibleAuthPlugin()]

    async def get_user(self, username: str) -> Optional[User]:
        return User(name=username, auth_plugin=GullibleAuthPlugin.name)



import io
import contextlib

from mysql_mimic import version


def test_version() -> None:
    assert isinstance(version.__version__, str)


def test_main() -> None:
    out = io.StringIO()
    with contextlib.redirect_stdout(out):
        version.main("__main__")

    assert f"{version.__version__}\n" == out.getvalue()
